# Copyright 2017-2021 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from itertools import groupby
from numbers import Real
import re
import pandas as pd
from pandas import DataFrame
from pandas.core.indexes.multi import MultiIndex

# Flake8 to ignore this import, it does not recognize that eval("np.nan") needs
# numpy package
import numpy as np  # noqa: F401

from .node import Node, traversal_order


class QueryMatcher:
    """Process and apply queries to GraphFrames."""

    def __init__(self, query=None):
        """Create a new QueryMatcher object.

        Arguments:
            query (list, optional): if provided, convert the contents of the high-level API query into an internal representation.
        """
        # Initialize containers for query and memoization cache.
        self.query_pattern = []
        self.search_cache = {}
        # If a high-level API list is provided, process it.
        if query is not None:
            assert isinstance(query, list)

            def _convert_dict_to_filter(attr_filter):
                """Converts high-level API attribute filter to a lambda"""
                compops = ("<", ">", "==", ">=", "<=", "<>", "!=")  # ,
                # Currently not supported
                #           "is", "is not", "in", "not in")

                # This is a dict to work around Python's non-local variable
                # assignment rules.
                #
                # TODO: Replace this with the use of the "nonlocal" keyword
                #       once Python 2.7 support is dropped.
                first_no_drop_indices = {"val": True}

                def filter_series(df_row):
                    def filter_single_series(df_row, key, single_value):
                        if key == "depth":
                            node = df_row.name
                            if isinstance(
                                single_value, str
                            ) and single_value.lower().startswith(compops):
                                return eval("{} {}".format(node._depth, single_value))
                            if isinstance(single_value, Real):
                                return node._depth == single_value
                            raise InvalidQueryFilter(
                                "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                                    key
                                )
                            )
                        if key == "node_id":
                            node = df_row.name
                            if isinstance(
                                single_value, str
                            ) and single_value.lower().startswith(compops):
                                return eval(
                                    "{} {}".format(node._hatchet_nid, single_value)
                                )
                            if isinstance(single_value, Real):
                                return node._hatchet_nid == single_value
                            raise InvalidQueryFilter(
                                "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                                    key
                                )
                            )
                        if key not in df_row.keys():
                            return False
                        if isinstance(df_row[key], str):
                            if not isinstance(single_value, str):
                                raise InvalidQueryFilter(
                                    "Value for attribute {} must be a string.".format(
                                        key
                                    )
                                )
                            return (
                                re.match(single_value + r"\Z", df_row[key]) is not None
                            )
                        if isinstance(df_row[key], Real):
                            if isinstance(
                                single_value, str
                            ) and single_value.lower().startswith(compops):
                                # compare nan metric value to numeric query
                                # (e.g. np.nan > 5)
                                if pd.isnull(df_row[key]):
                                    nan_str = "np.nan"
                                    # compare nan metric value to nan query
                                    # (e.g., np.nan == np.nan)
                                    if nan_str in single_value:
                                        return eval(
                                            "pd.isnull({}) == True".format(nan_str)
                                        )
                                    return eval("{} {}".format(nan_str, single_value))
                                elif np.isinf(df_row[key]):
                                    inf_str = "np.inf"
                                    # compare inf metric value to inf query
                                    # (e.g., np.inf == np.inf)
                                    if inf_str in single_value:
                                        return eval(
                                            "np.isinf({}) == True".format(inf_str)
                                        )
                                    return eval("{} {}".format(inf_str, single_value))
                                else:
                                    return eval(
                                        "{} {}".format(df_row[key], single_value)
                                    )

                            if isinstance(single_value, Real):
                                return df_row[key] == single_value
                            raise InvalidQueryFilter(
                                "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                                    key
                                )
                            )
                        raise InvalidQueryFilter(
                            "Filter must be one of the following:\n  * A regex string for a String attribute\n  * A string starting with a comparison operator for a Numeric attribute\n  * A number for a Numeric attribute\n"
                        )

                    matches = True
                    for k, v in attr_filter.items():
                        try:
                            _ = iter(v)
                            # Manually raise TypeError if v is a string so that
                            # the string is processed as a non-iterable
                            if isinstance(v, str):
                                raise TypeError
                        # Runs if v is not iterable (e.g., list, tuple, etc.)
                        except TypeError:
                            matches = matches and filter_single_series(df_row, k, v)
                        else:
                            for single_value in v:
                                matches = matches and filter_single_series(
                                    df_row, k, single_value
                                )
                    return matches

                def filter_dframe(df_row):
                    if first_no_drop_indices["val"]:
                        print(
                            "==================================================================="
                        )
                        print(
                            "WARNING: You are performing a query without dropping index levels."
                        )
                        print(
                            "         This is a valid operation, but it will significantly"
                        )
                        print(
                            "         increase the time it takes for this operation to complete."
                        )
                        print(
                            "         If you don't want the operation to take so long, call"
                        )
                        print("         GraphFrame.drop_index_levels() before calling")
                        print("         GraphFrame.filter()")
                        print(
                            "===================================================================\n"
                        )
                        first_no_drop_indices["val"] = False

                    def filter_single_dframe(node, df_row, key, single_value):
                        if key == "depth":
                            if isinstance(
                                single_value, str
                            ) and single_value.lower().startswith(compops):
                                return eval("{} {}".format(node._depth, single_value))
                            if isinstance(single_value, Real):
                                return node._depth == single_value
                            raise InvalidQueryFilter(
                                "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                                    key
                                )
                            )
                        if key == "node_id":
                            if isinstance(
                                single_value, str
                            ) and single_value.lower().startswith(compops):
                                return eval(
                                    "{} {}".format(node._hatchet_nid, single_value)
                                )
                            if isinstance(single_value, Real):
                                return node._hatchet_nid == single_value
                            raise InvalidQueryFilter(
                                "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                                    key
                                )
                            )
                        if key not in df_row.columns:
                            return False
                        if df_row[key].apply(type).eq(str).all():
                            if not isinstance(single_value, str):
                                raise InvalidQueryFilter(
                                    "Value for attribute {} must be a string.".format(
                                        key
                                    )
                                )
                            return (
                                df_row[key]
                                .apply(
                                    lambda x: re.match(single_value + r"\Z", x)
                                    is not None
                                )
                                .any()
                            )
                        if df_row[key].apply(type).eq(Real).all():
                            if isinstance(
                                single_value, str
                            ) and single_value.lower().startswith(compops):
                                return (
                                    df_row[key]
                                    .apply(
                                        lambda x: eval("{} {}".format(x, single_value))
                                    )
                                    .any()
                                )
                            if isinstance(single_value, Real):
                                return (
                                    df_row[key].apply(lambda x: x == single_value).any()
                                )
                            raise InvalidQueryFilter(
                                "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                                    key
                                )
                            )
                        raise InvalidQueryFilter(
                            "Filter must be one of the following:\n  * A regex string for a String attribute\n  * A string starting with a comparison operator for a Numeric attribute\n  * A number for a Numeric attribute\n"
                        )

                    matches = True
                    node = df_row.name.to_frame().index[0][0]
                    for k, v in attr_filter.items():
                        try:
                            _ = iter(v)
                            if isinstance(v, str):
                                raise TypeError
                        except TypeError:
                            matches = matches and filter_single_dframe(
                                node, df_row, k, v
                            )
                        else:
                            for single_value in v:
                                matches = matches and filter_single_dframe(
                                    node, df_row, k, single_value
                                )
                    return matches

                def filter_choice(df_row):
                    if isinstance(df_row, DataFrame):
                        return filter_dframe(df_row)
                    return filter_series(df_row)

                return filter_choice if attr_filter != {} else lambda row: True

            for elem in query:
                if isinstance(elem, dict):
                    self._add_node(".", _convert_dict_to_filter(elem))
                elif isinstance(elem, str) or isinstance(elem, int):
                    self._add_node(elem)
                elif isinstance(elem, tuple):
                    assert isinstance(elem[1], dict)
                    if isinstance(elem[0], str) or isinstance(elem[0], int):
                        self._add_node(elem[0], _convert_dict_to_filter(elem[1]))
                    else:
                        raise InvalidQueryPath(
                            "The first value of a tuple entry in a path must be either a string or integer."
                        )
                else:
                    raise InvalidQueryPath(
                        "A query path must be a list containing String, Integer, Dict, or Tuple elements"
                    )

    def match(self, wildcard_spec=".", filter_func=lambda row: True):
        """Start a query with a root node described by the arguments.

        Arguments:
            wildcard_spec (str, optional, ".", "*", or "+"): the wildcard status of the node (follows standard Regex syntax)
            filter_func (callable, optional): a callable accepting only a row from a Pandas DataFrame that is used to filter this node in the query

        Returns:
            (QueryMatcher): The instance of the class that called this function (enables fluent design).
        """
        if len(self.query_pattern) != 0:
            self.query_pattern = []
        self._add_node(wildcard_spec, filter_func)
        return self

    def rel(self, wildcard_spec=".", filter_func=lambda row: True):
        """Add another edge and node to the query.

        Arguments:
            wildcard_spec (str, optional, ".", "*", or "+"): the wildcard status of the node (follows standard Regex syntax)
            filter_func (callable, optional): a callable accepting only a row from a Pandas DataFrame that is used to filter this node in the query

        Returns:
            (QueryMatcher): The instance of the class that called this function (enables fluent design).
        """
        self._add_node(wildcard_spec, filter_func)
        return self

    def apply(self, gf):
        """Apply the query to a GraphFrame.

        Arguments:
            gf (GraphFrame): the GraphFrame on which to apply the query.

        Returns:
            (list): A list of lists representing the set of paths that match this query.
        """
        self.search_cache = {}
        matches = []
        visited = set()
        for root in sorted(gf.graph.roots, key=traversal_order):
            self._apply_impl(gf, root, visited, matches)
        assert len(visited) == len(gf.graph)
        return matches

    def _add_node(self, wildcard_spec=".", filter_func=lambda row: True):
        """Add a node to the query.
        Arguments:
            wildcard_spec (str, optional, ".", "*", or "+"): the wildcard status of the node (follows standard Regex syntax)
            filter_func (callable, optional): a callable accepting only a row from a Pandas DataFrame that is used to filter this node in the query
        """
        assert isinstance(wildcard_spec, int) or isinstance(wildcard_spec, str)
        assert callable(filter_func)
        if isinstance(wildcard_spec, int):
            for i in range(wildcard_spec):
                self.query_pattern.append((".", filter_func))
        else:
            assert wildcard_spec == "." or wildcard_spec == "*" or wildcard_spec == "+"
            self.query_pattern.append((wildcard_spec, filter_func))

    def _cache_node(self, gf, node):
        """Cache (Memoize) the parts of the query that the node matches.

        Arguments:
            gf (GraphFrame): the GraphFrame containing the node to be cached.
            node (Node): the Node to be cached.
        """
        assert isinstance(node, Node)
        matches = []
        # Applies each filtering function to the node to cache which
        # query nodes the current node matches.
        for i, node_query in enumerate(self.query_pattern):
            _, filter_func = node_query
            row = None
            if isinstance(gf.dataframe.index, MultiIndex):
                row = pd.concat([gf.dataframe.loc[node]], keys=[node], names=["node"])
            else:
                row = gf.dataframe.loc[node]
            if filter_func(row):
                matches.append(i)
        self.search_cache[node._hatchet_nid] = matches

    def _match_0_or_more(self, gf, node, wcard_idx):
        """Process a "*" wildcard in the query on a subgraph.

        Arguments:
            gf (GraphFrame): the GraphFrame being queried.
            node (Node): the node being queried against the "*" wildcard.
            wcard_idx (int): the index associated with the "*" wildcard query.

        Returns:
            (list): a list of lists representing the paths rooted at "node" that match the "*" wildcard and/or the next query node. Will return None if there is no match for the "*" wildcard or the next query node.
        """
        # Cache the node if it's not already cached
        if node._hatchet_nid not in self.search_cache:
            self._cache_node(gf, node)
        # If the node matches with the next non-wildcard query node,
        # end the recursion and return the node.
        if wcard_idx + 1 in self.search_cache[node._hatchet_nid]:
            return [[]]
        # If the node matches the "*" wildcard query, recursively
        # apply this function to the current node's children. Then,
        # collect their returned matches, and prepend the current node.
        elif wcard_idx in self.search_cache[node._hatchet_nid]:
            matches = []
            if len(node.children) == 0:
                if wcard_idx == len(self.query_pattern) - 1:
                    return [[node]]
                return None
            for child in sorted(node.children, key=traversal_order):
                sub_match = self._match_0_or_more(gf, child, wcard_idx)
                if sub_match is not None:
                    matches.extend(sub_match)
            if len(matches) == 0:
                return None
            tmp = set(tuple(m) for m in matches)
            matches = [list(t) for t in tmp]
            return [[node] + m for m in matches]
        # If the current node doesn't match the current "*" wildcard or
        # the next non-wildcard query node, return None.
        else:
            if wcard_idx == len(self.query_pattern) - 1:
                return [[]]
            return None

    def _match_1_or_more(self, gf, node, wcard_idx):
        """Process a "+" wildcard in the query on a subgraph.

        Arguments:
            gf (GraphFrame): the GraphFrame being queried.
            node (Node): the node being queried against the "+" wildcard.
            wcard_idx (int): the index associated with the "+" wildcard query.

        Returns:
            (list): a list of lists representing the paths rooted at "node" that match the "+" wildcard and/or the next query node. Will return None if there is no match for the "+" wildcard or the next query node.
        """
        # Cache the node if it's not already cached
        if node._hatchet_nid not in self.search_cache:
            self._cache_node(gf, node)
        # If the current node doesn't match the "+" wildcard, return None.
        if wcard_idx not in self.search_cache[node._hatchet_nid]:
            return None
        # Since a query can't end on a wildcard, return None if the
        # current node has no children.
        if len(node.children) == 0:
            return None
        # Use _match_0_or_more to collect all additional wildcard matches.
        matches = []
        for child in sorted(node.children, key=traversal_order):
            sub_match = self._match_0_or_more(gf, child, wcard_idx)
            if sub_match is not None:
                matches.extend(sub_match)
        # Since _match_0_or_more will capture the query node that follows
        # the wildcard, if no paths were retrieved from that function,
        # the pattern does not continue after the "+" wildcard. Thus,
        # since a pattern cannot end on a wildcard, the function
        # returns None.
        if len(matches) == 0:
            return None
        return [[node] + m for m in matches]

    def _match_1(self, gf, node, idx):
        """Process a "." wildcard in the query on a subgraph.

        Arguments:
            gf (GraphFrame): the GraphFrame being queried.
            node (Node): the node being queried against the "." wildcard.
            wcard_idx (int): the index associated with the "." wildcard query.

        Returns:
            (list): A list of lists representing the children of "node" that match the "." wildcard being considered. Will return None if there are no matches for the "." wildcard.
        """
        if node._hatchet_nid not in self.search_cache:
            self._cache_node(gf, node)
        matches = []
        for child in sorted(node.children, key=traversal_order):
            # Cache the node if it's not already cached
            if child._hatchet_nid not in self.search_cache:
                self._cache_node(gf, child)
            if idx in self.search_cache[child._hatchet_nid]:
                matches.append([child])
        # To be consistent with the other matching functions, return
        # None instead of an empty list.
        if len(matches) == 0:
            return None
        return matches

    def _match_pattern(self, gf, pattern_root, match_idx):
        """Try to match the query pattern starting at the provided root node.

        Arguments:
            gf (GraphFrame): the GraphFrame being queried.
            pattern_root (Node): the root node of the subgraph that is being queried.

        Returns:
            (list): A list of lists representing the paths rooted at "pattern_root" that match the query.
        """
        assert isinstance(pattern_root, Node)
        # Starting query node
        pattern_idx = match_idx + 1
        if (
            self.query_pattern[match_idx][0] == "*"
            or self.query_pattern[match_idx][0] == "+"
        ):
            pattern_idx = 0
        # Starting matching pattern
        matches = [[pattern_root]]
        while pattern_idx < len(self.query_pattern):
            # Get the wildcard type
            wcard, _ = self.query_pattern[pattern_idx]
            new_matches = []
            # Consider each existing match individually so that more
            # nodes can be added to them.
            for m in matches:
                sub_match = []
                # Get the portion of the subgraph that matches the next
                # part of the query.
                if wcard == ".":
                    s = self._match_1(gf, m[-1], pattern_idx)
                    if s is None:
                        sub_match.append(s)
                    else:
                        sub_match.extend(s)
                elif wcard == "*":
                    if len(m[-1].children) == 0:
                        sub_match.append([])
                    else:
                        for child in sorted(m[-1].children, key=traversal_order):
                            s = self._match_0_or_more(gf, child, pattern_idx)
                            if s is None:
                                sub_match.append(s)
                            else:
                                sub_match.extend(s)
                elif wcard == "+":
                    if len(m[-1].children) == 0:
                        sub_match.append(None)
                    else:
                        for child in sorted(m[-1].children, key=traversal_order):
                            s = self._match_1_or_more(gf, child, pattern_idx)
                            if s is None:
                                sub_match.append(s)
                            else:
                                sub_match.extend(s)
                else:
                    raise InvalidQueryFilter(
                        'Query wildcards must be one of ".", "*", or "+"'
                    )
                # Merge the next part of the match path with the
                # existing part.
                for s in sub_match:
                    if s is not None:
                        new_matches.append(m + s)
                new_matches = [uniq_match for uniq_match, _ in groupby(new_matches)]
            # Overwrite the old matches with the updated matches
            matches = new_matches
            # If all the existing partial matches were not able to be
            # expanded into full matches, return None.
            if len(matches) == 0:
                return None
            # Update the query node
            pattern_idx += 1
        return matches

    def _apply_impl(self, gf, node, visited, matches):
        """Traverse the subgraph with the specified root, and collect all paths that match the query.

        Arguments:
            gf (GraphFrame): the GraphFrame being queried.
            node (Node): the root node of the subgraph that is being queried.
            visited (set): a set that keeps track of what nodes have been visited in the traversal to minimize the amount of work that is repeated.
            matches (list): the list in which the final set of matches are stored.
        """
        # If the node has already been visited (or is None for some
        # reason), skip it.
        if node is None or node._hatchet_nid in visited:
            return
        # Cache the node if it's not already cached
        if node._hatchet_nid not in self.search_cache:
            self._cache_node(gf, node)
        # If the node matches the starting/root node of the query,
        # try to get all query matches in the subgraph rooted at
        # this node.
        if self.query_pattern[0][0] == "*":
            if 1 in self.search_cache[node._hatchet_nid]:
                sub_match = self._match_pattern(gf, node, 1)
                if sub_match is not None:
                    matches.extend(sub_match)
        if 0 in self.search_cache[node._hatchet_nid]:
            sub_match = self._match_pattern(gf, node, 0)
            if sub_match is not None:
                matches.extend(sub_match)
        # Note that the node is now visited.
        visited.add(node._hatchet_nid)
        # Continue the Depth First Search.
        for child in sorted(node.children, key=traversal_order):
            self._apply_impl(gf, child, visited, matches)


class InvalidQueryPath(Exception):
    """Raised when a query does not have the correct syntax"""


class InvalidQueryFilter(Exception):
    """Raised when a query filter does not have a valid syntax"""



#!/usr/bin/env python

#
# the list of components in timemory
#   see <timemory/components/types.hpp>
#
components = [
    "wall_clock",
    "system_clock",
    "user_clock",
    "cpu_clock",
    "monotonic_clock",
    "monotonic_raw_clock",
    "thread_cpu_clock",
    "process_cpu_clock",
    "cpu_util",
    "process_cpu_util",
    "thread_cpu_util",
    "peak_rss",
    "page_rss",
    "stack_rss",
    "data_rss",
    "num_swap",
    "num_io_in",
    "num_io_out",
    "num_minor_page_faults",
    "num_major_page_faults",
    "num_msg_sent",
    "num_msg_recv",
    "num_signals",
    "voluntary_context_switch",
    "priority_context_switch",
    "cuda_event",
    "cuda_profiler",
    "papi_array_t",
    "papi_vector",
    "caliper",
    "trip_count",
    "read_bytes",
    "written_bytes",
    "cupti_counters",
    "cupti_activity",
    "nvtx_marker",
    "cpu_roofline_sp_flops",
    "cpu_roofline_dp_flops",
    "cpu_roofline_flops",
    "gpu_roofline_hp_flops",
    "gpu_roofline_sp_flops",
    "gpu_roofline_dp_flops",
    "gpu_roofline_flops",
    "gperftools_cpu_profiler",
    "gperftools_heap_profiler",
    "virtual_memory",
    "likwid_marker",
    "likwid_nvmarker",
    "vtune_frame",
    "vtune_event",
    "vtune_profiler",
    "user_tuple_bundle",
    "user_list_bundle",
    "user_global_bundle",
    "tau_marker",
    "user_mode_time",
    "kernel_mode_time",
    "current_peak_rss",
    "malloc_gotcha",
    "user_mpip_bundle",
    "user_ompt_bundle",
    "ompt_handle",
    "allinea_map",
    "craypat_record",
    "craypat_region",
    "craypat_counters",
    "craypat_flush_buffer",
    "craypat_heap_stats"
]

#
# dictionary of components that have a different component
# name than the enumeration
# e.g. "component_name" : "enumeration_name"
#
mangled_enums = {
    "system_clock": "sys_clock",
    "papi_array_t": "papi_array",
}

#
# dictionary of components that have a different string
# identifier than than enumeration
# e.g. "component_name" : "string_identifier"
#
mangled_strings = {
    "wall_clock": ["real_clock", "virtual_clock"],
    "system_clock": ["sys_clock"],
    "papi_array_t": ["papi_array"],
    "papi_vector": ["papi"],
    "cpu_roofline_flops": ["cpu_roofline"],
    "gpu_roofline_flops": ["gpu_roofline"],
    "cpu_roofline_sp_flops": ["cpu_roofline_sp", "cpu_roofline_single"],
    "cpu_roofline_dp_flops": ["cpu_roofline_dp", "cpu_roofline_double"],
    "gpu_roofline_sp_flops": ["gpu_roofline_sp", "gpu_roofline_single"],
    "gpu_roofline_dp_flops": ["gpu_roofline_dp", "gpu_roofline_double"],
    "gpu_roofline_hp_flops": ["gpu_roofline_hp", "gpu_roofline_half"],
    "caliper": ["cali"],
    "written_bytes": ["write_bytes"],
    "nvtx_marker": ["nvtx"],
    "tau_marker": ["tau"],
    "gperf_cpu_profiler": ["gperf_cpu", "gperftools-cpu"],
    "gperf_heap_profiler": ["gperf_heap", "gperftools-heap"],
    "ompt_handle": ["ompt", "omp_tools", "openmp", "openmp_tools"],
    "mpip": ["mpi_tools", "mpi"],
    "user_global_bundle": ["global_bundle"],
    "user_list_bundle": ["list_bundle"],
    "user_tuple_bundle": ["tuple_bundle"],
}

recommended_types = {
    "tuple": ["wall_clock", "system_clock", "user_clock", "cpu_util",
              "page_rss", "peak_rss", "read_bytes", "written_bytes",
              "num_minor_page_faults", "num_major_page_faults",
              "voluntary_context_switch", "priority_context_switch"],
    "list": ["caliper", "papi_vector",
             "cuda_event", "nvtx_marker",
             "cupti_counters", "cupti_activity",
             "cpu_roofline_flops", "gpu_roofline_flops",
             "gperf_cpu_profiler", "gperf_heap_profiler"],
}

traits = {
    "is_timing_category": ("std::true_type",
                           [
                               "wall_clock",
                               "system_clock",
                               "user_clock",
                               "cpu_clock",
                               "monotonic_clock",
                               "monotonic_raw_clock",
                               "thread_cpu_clock",
                               "process_cpu_clock",
                               "cuda_event",
                               "cupti_activity",
                           ]),
    "is_memory_category": ("std::true_type",
                           [
                               "peak_rss",
                               "page_rss",
                               "stack_rss",
                               "data_rss",
                               "num_swap",
                               "num_io_in",
                               "num_io_out",
                               "num_minor_page_faults",
                               "num_major_page_faults",
                               "num_msg_sent",
                               "num_msg_recv",
                               "num_signals",
                               "voluntary_context_switch",
                               "priority_context_switch",
                               "read_bytes",
                               "written_bytes",
                               "virtual_memory",
                           ]),
    "uses_timing_units": ("std::true_type",
                          [
                              "wall_clock",
                              "system_clock",
                              "user_clock",
                              "cpu_clock",
                              "monotonic_clock",
                              "monotonic_raw_clock",
                              "thread_cpu_clock",
                              "process_cpu_clock",
                              "cuda_event",
                              "cupti_activity",
                          ]),
    "uses_memory_units": ("std::true_type",
                          [
                              "peak_rss",
                              "page_rss",
                              "stack_rss",
                              "data_rss",
                              "read_bytes",
                              "written_bytes",
                              "virtual_memory",
                          ]),
}

native_components = [
    "wall_clock",
    "system_clock",
    "user_clock",
    "cpu_clock",
    "monotonic_clock",
    "monotonic_raw_clock",
    "thread_cpu_clock",
    "process_cpu_clock",
    "cpu_util",
    "process_cpu_util",
    "thread_cpu_util",
    "peak_rss",
    "page_rss",
    "stack_rss",
    "data_rss",
    "num_swap",
    "num_io_in",
    "num_io_out",
    "num_minor_page_faults",
    "num_major_page_faults",
    "num_msg_sent",
    "num_msg_recv",
    "num_signals",
    "voluntary_context_switch",
    "priority_context_switch",
    "trip_count",
    "read_bytes",
    "written_bytes",
    "virtual_memory",
]



# Template service tests

import unittest

import calipertest as calitest

class CaliperTemplateServicesTest(unittest.TestCase):
    """ Caliper template services test cases """

    def test_measurement_template(self):
        target_cmd = [ './ci_test_basic' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'event,measurement_template,trace,recorder',
            'CALI_MEASUREMENT_TEMPLATE_NAMES' : 'ci_test',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = calitest.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = calitest.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) > 1)

        self.assertTrue(calitest.has_snapshot_with_keys(
            snapshots, { 'phase',
                         'measurement.val.ci_test',
                         'measurement.ci_test' }))

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
#
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import
from __future__ import division

import os
import sys
import argparse

__author__ = "Jonathan Madsen"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Jonathan Madsen"]
__license__ = "MIT"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"

from .analyze import (
    load,
    match,
    search,
    expression,
    sort,
    group,
    add,
    subtract,
    unify,
    dump,
)


def embedded_analyze(
    args=None,
    data=[],
    call_exit=False,
    verbose=os.environ.get("TIMEMORY_VERBOSE", 0),
    ranks=[],
):
    """This is intended to be called from the embedded python interpreter"""
    if len(ranks) > 0:
        try:
            import mpi4py  # noqa: F401
            from mpi4py import MPI  # noqa: F401

            rank = MPI.COMM_WORLD.Get_rank()
            if rank not in ranks:
                return
        except (ImportError, RuntimeError):
            pass

    call_exit = False
    cmd_line = False
    if args is None:
        args = sys.argv[:]
        call_exit = True
        cmd_line = True

    try:
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "files",
            metavar="file",
            type=str,
            nargs="*" if not cmd_line else "+",
            help="Files to analyze",
            default=[],
        )
        parser.add_argument(
            "-f",
            "--format",
            type=str,
            help="Data output format.",
            choices=(
                "dot",
                "flamegraph",
                "tree",
                "table",
                "markdown",
                "html",
                "markdown_grid",
            ),
            nargs="*",
            default=None,
        )
        parser.add_argument(
            "-o",
            "--output",
            type=str,
            help="Output to file(s)",
            nargs="*",
            default=None,
        )
        parser.add_argument(
            "-M",
            "--mode",
            help="Analysis mode",
            nargs="*",
            choices=("add", "subtract", "unify", "group"),
            default=[],
        )
        parser.add_argument(
            "-m",
            "--metric",
            "--column",
            type=str,
            help="Metric(s) to extract",
            default="sum.inc",
        )
        parser.add_argument(
            "-s",
            "--sort",
            type=str,
            help="Sort the metric",
            choices=("ascending", "descending"),
            default=None,
        )
        parser.add_argument(
            "-g",
            "--group",
            type=str,
            help="Group by a column name in the dataframe",
            default=None,
        )
        parser.add_argument(
            "--field",
            type=str,
            help="Dataframe column to search/match against",
            default="name",
        )
        parser.add_argument(
            "--search",
            type=str,
            default=None,
            help="Regular expression for re.search(...), i.e. a substring match",
        )
        parser.add_argument(
            "--match",
            type=str,
            default=None,
            help="Regular expression for re.match(...), i.e. a full string match",
        )
        parser.add_argument(
            "--expression",
            type=str,
            default=None,
            help=(
                "A space-delimited comparison operation expression using 'x' "
                + "for the variable, numerical values, and: < <= > >= && ||. "
                + "E.g. 'x > 1.0e3 && x < 100000'. 'x' will be -m/--metric"
            ),
        )
        parser.add_argument(
            "-e",
            "--echo-dart",
            help="echo Dart measurement for CDash",
            action="store_true",
        )
        parser.add_argument(
            "--per-thread",
            help=(
                "Encode the thread ID in node hash to ensure squashing doesn't "
                + "combine thread-data"
            ),
            action="store_true",
        )
        parser.add_argument(
            "--per-rank",
            help=(
                "Encode the rank ID in node hash to ensure squashing doesn't "
                + "combine rank-data"
            ),
            action="store_true",
        )
        parser.add_argument(
            "--select",
            help=(
                "Select the component type if the JSON input contains output "
                + "from multiple components"
            ),
            type=str,
            default=None,
        )
        parser.add_argument(
            "--exit-on-failure",
            help="Abort with non-zero exit code if errors arise",
            action="store_true",
        )

        _args = parser.parse_args(args)

        if _args.exit_on_failure:
            call_exit = True

        if _args.group is None and _args.format is None:
            _args.format = ["tree"]
        if _args.group is not None and _args.format not in ("table", None):
            raise RuntimeError("Invalid data format for group")

        if isinstance(_args.select, str):
            _args.select = _args.select.split()

        gfs = []
        for itr in data:
            if not isinstance(itr, dict):
                print("data: {}\ndata-type: {}".format(itr, type(itr).__name__))
            gfs.append(
                load(
                    itr,
                    select=_args.select,
                    per_thread=_args.per_thread,
                    per_rank=_args.per_rank,
                )
            )
        for itr in _args.files:
            if not os.path.exists(itr):
                print("file: {}\nfile-type: {}".format(itr, type(itr).__name__))
            gfs.append(
                load(
                    itr,
                    select=_args.select,
                    per_thread=_args.per_thread,
                    per_rank=_args.per_rank,
                )
            )

        def apply(_input, _mode, *_args, **_kwargs):
            if _mode == "search":
                return search(_input, *_args, **_kwargs)
            elif _mode == "match":
                return match(_input, *_args, **_kwargs)
            elif _mode == "expression":
                return expression(_input, *_args, **_kwargs)
            elif _mode == "sort":
                return sort(_input, *_args, **_kwargs)
            elif _mode == "add":
                return add(_input)
            elif _mode == "subtract":
                return subtract(_input)
            elif _mode == "unify":
                return unify(_input)
            elif _mode == "group":
                return group(_input, *_args, **_kwargs)

        # apply search before match since this is less restrictive
        if _args.search is not None:
            gfs = apply(gfs, "search", pattern=_args.search, field=_args.field)

        # apply match after search since this is more restrictive
        if _args.match is not None:
            gfs = apply(gfs, "match", pattern=_args.match, field=_args.field)

        # apply numerical expression last
        if _args.expression is not None:
            gfs = apply(
                gfs,
                "expression",
                math_expr=_args.expression,
                metric=_args.metric,
            )

        # apply the mutating operations
        if "add" in _args.mode:
            gfs = apply(gfs, "add")
        elif "subtract" in _args.mode:
            gfs = apply(gfs, "subtract")
        elif "unify" in _args.mode:
            gfs = apply(gfs, "unify")

        if _args.sort is not None:
            gfs = apply(
                gfs,
                "sort",
                metric=_args.metric,
                ascending=(_args.sort == "ascending"),
            )

        files = _args.output
        if files is not None and len(files) == 1:
            files = files[0]

        if _args.group is not None:
            gfs = apply(
                gfs,
                "group",
                metric=_args.metric,
                field=_args.group,
                ascending=(_args.sort == "ascending"),
            )
            _args.format = [None]

        for fmt in _args.format:
            dump(gfs, _args.metric, fmt, files, _args.echo_dart)

    except Exception as e:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        import traceback

        traceback.print_exception(exc_type, exc_value, exc_traceback, limit=10)
        print("Exception - {}".format(e))
        if call_exit or _args.exit_on_failure:
            sys.exit(1)
        elif not cmd_line:
            raise



#!@PYTHON_EXECUTABLE@
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import

__author__ = "Jonathan Madsen"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Jonathan Madsen"]
__license__ = "MIT"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"

try:
    import mpi4py  # noqa: F401
    from mpi4py import MPI  # noqa: F401
except ImportError:
    pass

import time
import unittest
import timemory as tim


# --------------------------- helper functions ----------------------------------------- #
# check availability of a component
def check_available(component):
    return getattr(component, "available")


# compute fibonacci
def fibonacci(n):
    return n if n < 2 else (fibonacci(n - 1) + fibonacci(n - 2))


# cpu utilization for n milliseconds
def consume(n):
    # Python 3.7 has time_ns()
    try:
        # get current time in nsec
        now = time.time_ns()
        # try until time point
        while time.time_ns() < (now + (n * 1e6)):
            pass
    except AttributeError:
        now = 1000 * time.time()
        # try until time point
        while (1000 * time.time()) < (now + n):
            pass


# --------------------------- General Tests set ---------------------------------------- #
# General tests class
class TimemoryGeneralTests(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # set up environment variables
        tim.settings.verbose = 1
        tim.settings.debug = False
        tim.settings.tree_output = False
        tim.settings.text_output = True
        tim.settings.cout_output = False
        tim.settings.json_output = False
        tim.settings.flamegraph_output = False
        tim.settings.mpi_thread = False
        tim.settings.dart_output = True
        tim.settings.dart_count = 1
        tim.settings.banner = False

    def setUp(self):
        pass

    def tearDown(self):
        pass

    @classmethod
    def tearDownClass(self):
        pass

    # ---------------------------------------------------------------------------------- #
    # test handling hash identifier access
    def test_hashing(self):
        """hashing"""

        fib_hash = tim.get_hash("fibonacci")
        fib_val = tim.add_hash_id("fibonacci")

        with tim.util.marker(["wall_clock"], key="fibonacci"):
            fibonacci(20)

        fib_id = tim.get_hash_identifier(fib_val)

        self.assertEqual(fib_hash, fib_val)
        self.assertEqual(fib_id, "fibonacci")

        foobar_hash = tim.get_hash("foobar")
        foobar_val = tim.add_hash_id("foobar")
        foobar_id = tim.get_hash_identifier(foobar_val)

        self.assertEqual(foobar_hash, foobar_val)
        self.assertEqual(foobar_id, "foobar")

        print(f"fib_hash    : {fib_hash}")
        print(f"fib_val     : {fib_val}")
        print(f"fib_id      : {fib_id}")
        print(f"foobar_hash : {foobar_hash}")
        print(f"foobar_val  : {foobar_val}")
        print(f"foobar_id   : {foobar_id}")


# ----------------------------- main test runner -------------------------------------- #
# main runner
def run():
    # run all tests
    unittest.main()


if __name__ == "__main__":
    tim.initialize([__file__])
    run()



# report / C config test

import json,unittest

import calipertest as cat

class CaliperReportTest(unittest.TestCase):
    """ Caliper Report / Reader CSV export / C config API test case """

    def test_report_caliwriter(self):
        """ Test reader lib's CSV export via report service """

        target_cmd = [ './ci_test_macros' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query',
                       '-q', 'SELECT count(),iteration#fooloop WHERE loop=fooloop GROUP BY iteration#fooloop FORMAT expand' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'event,trace,report',
            'CALI_REPORT_CONFIG'     : 'format cali',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) == 5)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'iteration#fooloop': '3', 'count': '4' }))

    def test_report_aggregate(self):
        """ Test reader lib's CSV export via report service """

        target_cmd = [ './ci_test_macros' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'event,aggregate,report',
            'CALI_REPORT_CONFIG'     : 'select function,count(),inclusive_sum(count) group by function format expand',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output,_ = cat.run_test(target_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function': 'main', 'count': '17', 'inclusive#count': '77' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function': 'main/foo', 'count': '60', 'inclusive#count': '60' }))

    def test_report_nested_key(self):
        target_cmd = [ './ci_test_macros' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'event,trace,report',
            'CALI_REPORT_CONFIG'     : 'select *,count() group by prop:nested,iteration#mainloop format expand',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output,_ = cat.run_test(target_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function'   : 'main/foo',
                         'loop'       : 'mainloop',
                         'annotation' : 'pre-loop',
                         'statement'  : 'foo.init',
                         'iteration#mainloop' : '2',
                         'count'      : '1'
            }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function'   : 'main/foo',
                         'loop'       : 'mainloop/fooloop',
                         'iteration#mainloop' : '3',
                         'count'      : '9'
            }))

    def test_report_class_iteration(self):
        target_cmd = [ './ci_test_macros' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'event,trace,report',
            'CALI_REPORT_CONFIG'     : 'select *,count() group by prop:nested,class.iteration format expand',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output,_ = cat.run_test(target_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function'   : 'main/foo',
                         'loop'       : 'mainloop/fooloop',
                         'iteration#mainloop' : '3',
                         'iteration#fooloop'  : '1',
                         'count'      : '1'
            }))

    def test_report_aggregation_aliases(self):
        target_cmd = [ './ci_test_macros' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'event,trace,report',
            'CALI_REPORT_CONFIG'     : 'select *,count() as my\\ count\\ alias group by prop:nested,iteration#mainloop format expand',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output,_ = cat.run_test(target_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function'   : 'main/foo',
                         'loop'       : 'mainloop',
                         'annotation' : 'pre-loop',
                         'statement'  : 'foo.init',
                         'iteration#mainloop' : '2',
                         'my count alias'     : '1'
            }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function'   : 'main/foo',
                         'loop'       : 'mainloop/fooloop',
                         'iteration#mainloop' : '3',
                         'my count alias'     : '9'
            }))

    def test_report_tree_formatter(self):
        target_cmd = [ './ci_test_macros' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'event,trace,report',
            'CALI_REPORT_CONFIG'     : 'select *,count() as Count group by prop:nested where cali.event.end format tree',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        report_targets = [
            'Path             Count',
            'main                 1',
            '  mainloop           5',
            '    foo              4',
            '      fooloop       20',
            '      pre-loop       4',
            '        foo.init     4',
            '  before_loop        1'
        ]

        report_output,_ = cat.run_test(target_cmd, caliper_config)
        lines = report_output.decode().splitlines()

        for target in report_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in report' % target)

    def test_report_table_formatter(self):
        target_cmd = [ './ci_test_macros' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'event,trace,report',
            'CALI_REPORT_CONFIG'     : 'select *,count() group by loop where event.end#loop format table order by count desc',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        report_targets = [
            'loop             count',
            'mainloop/fooloop     4',
            'mainloop             1'
        ]

        report_output,_ = cat.run_test(target_cmd, caliper_config)
        output_lines = report_output.decode().splitlines()

        for lineno in [ 0, 1, 2]:
            self.assertEqual(report_targets[lineno], output_lines[lineno].rstrip())

    def test_report_attribute_aliases(self):
        target_cmd = [ './ci_test_macros' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'event,trace,report',
            'CALI_REPORT_CONFIG'     : 'select function as my\\ function\\ alias,count() group by function format expand',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output,_ = cat.run_test(target_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(cat.has_snapshot_with_keys(snapshots, { 'my function alias', 'count'  }))

    def test_report_attribute_aliases_in_cali(self):
        """ Test json object layout """

        target_cmd = [ './ci_test_macros' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-q', 'format json(object)' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'event,trace,report',
            'CALI_REPORT_FILENAME'   : 'stdout',
            'CALI_REPORT_CONFIG'     : 'select *,count() as CountAlias group by prop:nested format cali',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        obj = json.loads( cat.run_test_with_query(target_cmd, query_cmd, caliper_config ) )

        self.assertTrue( { 'records', 'globals', 'attributes' }.issubset(set(obj.keys())) )

        self.assertIn('count',           obj['attributes'])
        self.assertIn('attribute.alias', obj['attributes']['count'])

        self.assertEqual('CountAlias', obj['attributes']['count']['attribute.alias'])

    def test_report(self):
        target_cmd = [ './ci_test_report' ]
        # create some distraction: read-from-env should be disabled
        env = { 'CALI_SERVICES_ENABLE' : 'debug' }

        report_out,_ = cat.run_test(target_cmd, env)

        lines = report_out.decode().splitlines()

        self.assertEqual(lines[0].rstrip(), 'function annotation count')
        self.assertEqual(lines[1].rstrip(), 'main     my phase       2')

if __name__ == "__main__":
    unittest.main()



from .python25 cimport PyFrameObject, PyObject, PyStringObject
from libc.stdlib cimport free

def _default_component():
    return None

_component_generator = _default_component
_component_display_units = "sec"
_component_num_prealloc = 0

def set_component_prealloc(_num_prealloc):
    global _component_num_prealloc
    _component_num_prealloc = _num_prealloc

def set_component_generator(_gen):
    global _component_generator
    global _component_display_units
    _component_generator = _gen
    try:
        _component_display_units = _gen().display_unit()
    except:
        pass

def get_component():
    global _component_generator
    return _component_generator()

cdef extern from "frameobject.h":
    ctypedef int (*Py_tracefunc)(object self, PyFrameObject *py_frame, int what, PyObject *arg)

cdef extern from "Python.h":
    ctypedef long long PY_LONG_LONG
    cdef bint PyCFunction_Check(object obj)

    cdef void PyEval_SetProfile(Py_tracefunc func, object arg)
    cdef void PyEval_SetTrace(Py_tracefunc func, object arg)

    ctypedef object (*PyCFunction)(object self, object args)

    ctypedef struct PyMethodDef:
        char *ml_name
        PyCFunction ml_meth
        int ml_flags
        char *ml_doc

    ctypedef struct PyCFunctionObject:
        PyMethodDef *m_ml
        PyObject *m_self
        PyObject *m_module

    # They're actually #defines, but whatever.
    cdef int PyTrace_CALL
    cdef int PyTrace_EXCEPTION
    cdef int PyTrace_LINE
    cdef int PyTrace_RETURN
    cdef int PyTrace_C_CALL
    cdef int PyTrace_C_EXCEPTION
    cdef int PyTrace_C_RETURN

cdef extern from "unset_trace.h":
    void unset_trace()

def label(code):
    """ Return a (filename, first_lineno, func_name) tuple for a given code
    object.

    This is the same labelling as used by the cProfile module in Python 2.5.
    """
    if isinstance(code, str):
        return ('~', 0, code)    # built-in functions ('~' sorts at the end)
    else:
        return (code.co_filename, code.co_firstlineno, code.co_name)


cdef class LineMetrics:
    """ The timing for a single line.
    """
    cdef public object code
    cdef public int lineno
    cdef public object comp

    def __init__(self, object code, int lineno, object comp = get_component()):
        self.code = code
        self.lineno = lineno
        self.comp = comp

    def valid(self):
        return (self.comp.laps() > 0)

    def start(self):
        self.comp.start()

    def stop(self):
        self.comp.stop()

    def astuple(self):
        """ Convert to a tuple of (lineno, hits, total).
        """
        return (self.lineno, self.comp.laps(), self.comp.get())

    def __repr__(self):
        return '<LineMetrics for %r\n  lineno: %r\n  hits: %r\n  total: %r>' % (self.code, self.lineno, self.comp.laps(), self.comp.get())


# Note: this is a regular Python class to allow easy pickling.
class LineStats(object):
    """ Object to encapsulate line-profile statistics.

    Attributes
    ----------
    metrics : dict
        Mapping from (filename, first_lineno, function_name) of the profiled
        function to a list of (lineno, nhits, total) tuples for each
        profiled line. total is an integer in the native units of the
        timer.
    unit : float
        The number of seconds per timer unit.
    """
    def __init__(self, metrics, unit, display_unit):
        self.metrics = metrics
        self.unit = unit
        self.display_unit = display_unit


cdef class LineProfiler:
    """ Time the execution of lines of Python code.
    """
    cdef public long last_index
    cdef public object last_tool
    cdef public object last_code
    #
    cdef public list functions
    cdef public dict code_map
    cdef public long enable_count
    cdef public double unit

    def __init__(self, *functions):
        global _component_display_units

        self.functions = []
        self.code_map = {}
        self.last_index = 0
        self.last_tool = None
        self.last_code = None
        self.unit = get_component().unit()
        self.display_unit = _component_display_units
        self.enable_count = 0
        for func in functions:
            self.add_function(func)

    def add_function(self, func):
        global _component_num_prealloc

        """ Record line profiling information for the given Python function.
        """
        try:
            code = func.__code__
        except AttributeError:
            import warnings
            warnings.warn("Could not extract a code object for the object %r" % (func,))
            return
        if code not in self.code_map:
            self.code_map[code] = {}
            self.functions.append(func)
        for i in range(_component_num_prealloc):
            self.code_map[code][i] = LineMetrics(code, i, get_component())
        if _component_num_prealloc > 0:
            print("Number of pre-allocated components for {}: {}".format(
                func.__name__, _component_num_prealloc))

    def enable_by_count(self):
        """ Enable the profiler if it hasn't been enabled before.
        """
        if self.enable_count == 0:
            self.enable()
        self.enable_count += 1

    def disable_by_count(self):
        """ Disable the profiler if the number of disable requests matches the
        number of enable requests.
        """
        if self.enable_count > 0:
            self.enable_count -= 1
            if self.enable_count == 0:
                self.disable()

    def __enter__(self):
        self.enable_by_count()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disable_by_count()

    def enable(self):
        PyEval_SetTrace(python_trace_callback, self)

    def disable(self):
        unset_trace()

    def cleanup(self):
        # cleanup any lines pre-allocated but without entries
        for code in self.code_map:
            rmkeys = []
            for f, s in self.code_map[code].items():
                if not s.valid():
                    rmkeys.append(f)
            for i in rmkeys:
                del self.code_map[code][i]

    def get_stats(self):
        """ Return a LineStats object containing the metrics.
        """
        self.cleanup()
        stats = {}
        for code in self.code_map:
            entries = self.code_map[code].values()
            key = label(code)
            stats[key] = [e.astuple() for e in entries]
            stats[key].sort()
        return LineStats(stats, self.unit, self.display_unit)


cdef int python_trace_callback(object self_, PyFrameObject *py_frame, int what,
    PyObject *arg):
    """ The PyEval_SetTrace() callback.
    """
    cdef LineProfiler self
    cdef long nline
    cdef object code
    cdef LineMetrics comp

    self = <LineProfiler>self_
    code_map = self.code_map
    if self.last_tool is not None:
        self.last_tool.stop()

    if what == PyTrace_LINE:
        code = <object>py_frame.f_code
        if code in code_map:
            # the current line
            nline = py_frame.f_lineno
            if nline not in code_map[code]:
                comp = LineMetrics(code, nline, get_component())
                code_map[code][nline] = comp
            else:
                comp = code_map[code][nline]
            self.last_tool = comp
            self.last_code = code
            self.last_index = nline
            comp.start()

    return 0





# Loop report tests

import json
import os
import unittest

import calipertest as cat

class CaliperLoopReportTest(unittest.TestCase):
    """ Loop report controller (summary info) """
    def test_loopreport_summary(self):
        target_cmd = [ './ci_test_macros', '10', 'loop-report,summary=true,timeseries=false,output=stdout', '20' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'      : '0',
        }

        log_targets = [
            'Loop summary:',
            'Loop     Iterations Time (s)',
            'mainloop         20'
        ]

        report_out,_ = cat.run_test(target_cmd, caliper_config)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

    """ Loop report controller (timeseries info) """
    def test_loopreport_timeseries(self):
        target_cmd = [ './ci_test_macros', '0', 'loop-report,timeseries=true,summary=false,iteration_interval=15,output=stdout', '75' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'      : '0',
        }

        log_targets = [
            'Iteration summary (mainloop):',
            'Block Iterations Time (s)',
            '    0         15',
            '   30         15',
            '   60         15'
        ]

        report_out,_ = cat.run_test(target_cmd, caliper_config)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

    """ Loop report controller (timeseries info, display default number of rows (20)) """
    def test_loopreport_display_default_num_rows(self):
        target_cmd = [ './ci_test_macros', '0', 'loop-report,timeseries=true,summary=false,iteration_interval=2,output=stdout', '80' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'      : '0',
        }

        report_out,_ = cat.run_test(target_cmd, caliper_config)
        lines = report_out.decode().splitlines()

        self.assertGreater(len(lines), 20)
        self.assertLess(len(lines), 28)

    """ Loop report controller (timeseries info, display all rows) """
    def test_loopreport_display_all_rows(self):
        target_cmd = [ './ci_test_macros', '0', 'loop-report,timeseries.maxrows=0,timeseries=true,summary=false,iteration_interval=2,output=stdout', '80' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'      : '0',
        }

        report_out,_ = cat.run_test(target_cmd, caliper_config)
        lines = report_out.decode().splitlines()

        self.assertGreater(len(lines), 40)
        self.assertLess(len(lines), 48)

    """ Loop report controller (timeseries info, target loop selection) """
    def test_loopreport_target_loop_selection(self):
        target_cmd = [ './ci_test_macros', '0', 'loop-report,target_loops=fooloop,timeseries=true,summary=true,iteration_interval=5,output=stdout,timeseries.maxrows=0', '20' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'      : '0',
        }

        log_targets = [
            'Iteration summary (fooloop):',
            'mainloop/fooloop        400',
            'Block Iterations Time (s)',
            '    0        100',
            '    5        100',
            '   10        100'
        ]

        report_out,_ = cat.run_test(target_cmd, caliper_config)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
#
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

# @file mpi_support.py
# MPI information for timemory module
#

from __future__ import absolute_import
import os

__all__ = [
    "mpi_exe_info",
    "mpi_c_info",
    "mpi_cxx_info",
    "is_supported",
    "enable_mpip",
]


def _path_to_file(pfile):
    return "{}/{}".format(os.path.dirname(__file__), pfile)


def mpi_exe_info():
    """
    Returns information about the MPI executable used in compilation
    """
    print(open(_path_to_file("mpi_exe_info.txt"), "r").read())


def mpi_c_info():
    """
    Returns information about the MPI C libraries used in compilation
    """
    print(open(_path_to_file("mpi_c_info.txt"), "r").read())


def mpi_cxx_info():
    """
    Returns information about the MPI C++ libraries used in compilation
    """
    print(open(_path_to_file("mpi_cxx_info.txt"), "r").read())


def is_supported():
    """
    Returns whether MPI is supported
    """
    import timemory

    return timemory.has_mpi_support()


def enable_mpip(permit=[], reject=[]):
    """
    Enable the MPIP Gotcha routines (Linux-only)

    Arguments:
        reject_list (list):   the names of the MPI functions to exclude from wrapping
    """
    import timemory

    if len(permit) > 0:
        os.environ["TIMEMORY_MPIP_PERMIT_LIST"] = ";".join(permit)
    if len(reject) > 0:
        os.environ["TIMEMORY_MPIP_REJECT_LIST"] = ";".join(reject)
    timemory.start_mpip()



#!@PYTHON_EXECUTABLE@
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import

__author__ = "Muhammad Haseeb"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Muhammad Haseeb"]
__license__ = "MIT"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"

try:
    import mpi4py  # noqa: F401
    from mpi4py import MPI  # noqa: F401
except ImportError:
    pass

import os
import time
import json
import random
import unittest
import timemory as tim
from timemory.profiler import profile, config
from timemory.bundle import marker

# --------------------------- test setup variables ----------------------------------- #


# --------------------------- helper functions ----------------------------------------- #
# compute fibonacci
def fibonacci(n):
    return n if n < 2 else (fibonacci(n - 1) + fibonacci(n - 2))


def fib(n, instr):
    if instr is True:
        with marker(components=["wall_clock", "current_peak_rss"], key="fib"):
            return n if n < 2 else (fib(n - 1, True) + fib(n - 2, False))
    else:
        return n if n < 2 else (fibonacci(n - 1) + fibonacci(n - 2))


# sleep for n millisec
def do_sleep(n):
    time.sleep(n * 1e-3)


# cpu utilization for n millisec
def consume(n):
    # Python 3.7 has time_ns()
    try:
        # get current time in nsec
        now = time.time_ns()
        # try until time point
        while time.time_ns() < (now + (n * 1e6)):
            pass
    except AttributeError:
        now = 1000 * time.time()
        # try until time point
        while (1000 * time.time()) < (now + n):
            pass


# get a random entry
def random_entry(vect):
    random.seed(int(1.0e9 * time.time()))
    num = int(random.uniform(0, len(vect) - 1))
    return vect[num]


# get auto_tuple config
def get_config(items=["wall_clock"]):
    return [getattr(tim.component, x) for x in items]


# -------------------------- Thottle Tests set ---------------------------------------- #
# Flat tests class
class TimemoryFlatTests(unittest.TestCase):
    # setup class: timemory settings
    @classmethod
    def setUpClass(self):
        # set up environment variables
        tim.settings.verbose = 1
        tim.settings.debug = False
        tim.settings.json_output = True
        tim.settings.mpi_thread = False
        tim.settings.dart_output = True
        tim.settings.dart_count = 1
        tim.settings.banner = False
        tim.settings.parse()

    def setUp(self):
        tim.settings.flat_profile = True
        config.include_internal = True

    def tearDown(self):
        config.include_internal = False

    # Tear down class: finalize
    @classmethod
    def tearDownClass(self):
        pass

    # ---------------------------------------------------------------------------------- #
    # test profiler_depth
    def test_parse(self):
        """parse"""

        tim.settings.flat_profile = True
        self.assertTrue(tim.settings.flat_profile)

        tim.settings.flat_profile = False
        self.assertFalse(tim.settings.flat_profile)

        os.environ["TIMEMORY_FLAT_PROFILE"] = "ON"
        tim.settings.parse()

        ret = os.environ.get("TIMEMORY_FLAT_PROFILE")

        self.assertTrue(ret)
        self.assertTrue(tim.settings.flat_profile)

        # reset
        tim.settings.flat_profile = False

    # ---------------------------------------------------------------------------------- #
    # test profiler_depth
    def test_no_flat(self):
        """not_flat"""
        n = 10
        tim.settings.flat_profile = False
        with marker(
            components=["wall_clock", "cpu_clock"], key=self.shortDescription()
        ):
            fib(n, True)

        print(
            "{}\n".format(
                json.dumps(tim.get()["timemory"]["wall_clock"], indent=4)
            )
        )
        # inspect data
        data = tim.get()["timemory"]["wall_clock"]["ranks"][0]["graph"]
        self.assertEqual(data[-1]["depth"], n)

    # ---------------------------------------------------------------------------------- #
    # test profiler_depth
    def test_flat(self):
        """flat"""
        n = 10
        with marker(
            components=["wall_clock", "cpu_clock"], key=self.shortDescription()
        ):
            with profile(
                components=["wall_clock", "cpu_clock"],
                flat=True,
                timeline=False,
            ):
                fibonacci(n)

        print(
            "{}\n".format(
                json.dumps(tim.get()["timemory"]["wall_clock"], indent=4)
            )
        )
        # inspect data
        data = tim.get()["timemory"]["wall_clock"]["ranks"][0]["graph"]
        self.assertEqual(data[-1]["depth"], 0)


# ----------------------------- main test runner -------------------------------------- #
# main runner
def run():
    # run all tests
    unittest.main()


if __name__ == "__main__":
    tim.initialize([__file__])
    run()



# Multi-channel tests

import unittest

import calipertest as cat

class CaliperMultiChannelTest(unittest.TestCase):
    """ Caliper multi-channel case """

    def test_multichannel_trace(self):
        target_cmd = [ './ci_test_multichannel' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_CONFIG_PROFILE'    : 'serial-trace',
            'CALI_CHANNEL_SNAPSHOT_SCOPES' : 'process,thread,channel',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) >= 205)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, {'chn.id'       : '1', 
                        'thread'       : 'true' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, {'chn.id'       : '1', 
                        'main'         : 'true' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, {'chn.id'       : '42', 
                        'thread'       : 'true' }))

    def test_multichannel_aggr(self):
        target_cmd = [ './ci_test_multichannel' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'aggregate,event,recorder',
            'CALI_CHANNEL_SNAPSHOT_SCOPES' : 'process,thread,channel',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) >= 210)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, {'chn.id'       : '1', 
                        'thread'       : 'true' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, {'chn.id'       : '1', 
                        'main'         : 'true' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, {'chn.id'       : '42', 
                        'thread'       : 'true' }))

    def test_channel_c_api(self):
        target_cmd = [ './ci_test_channel_api' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY' : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'annotation': 'foo',
                         'a': '2',
                         'b': '4' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'annotation': 'foo',
                         'b': '4',
                         'c': '8' }))
        self.assertFalse(cat.has_snapshot_with_attributes(
            snapshots, { 'annotation': 'foo',
                         'a': '2',
                         'b': '4',
                         'c': '8' }))
        

if __name__ == "__main__":
    unittest.main()



# Test the loop and region monitor services

import json
import unittest

import calipertest as cat

class CaliperTestMonitor(unittest.TestCase):
    """ Caliper test case """

    def test_loopmonitor_iter_interval(self):
        target_cmd = [ './ci_test_macros', '0', 'none', '10' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'loop_monitor,trace,report',
            'CALI_LOOP_MONITOR_ITERATION_INTERVAL' : '5',
            'CALI_REPORT_CONFIG'     : 'select * where iteration#mainloop format expand',
            'CALI_REPORT_FILENAME'   : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test(target_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output[0])

        self.assertTrue(len(snapshots) == 2)
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'iteration#mainloop' : '4',
                         'loop.iterations'    : '5'
            }))

    def test_loopmonitor_time_interval(self):
        target_cmd = [ './ci_test_macros', '2000', 'none', '10' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'loop_monitor,trace,report',
            'CALI_LOOP_MONITOR_TIME_INTERVAL' : '0.05',
            'CALI_REPORT_CONFIG'     : 'select count(),sum(loop.iterations) where loop format json',
            'CALI_REPORT_FILENAME'   : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test(target_cmd, caliper_config)
        obj = json.loads( query_output[0] )

        self.assertTrue(len(obj) == 1)

        count = int(obj[0]['count'])
        sum   = int(obj[0]['sum#loop.iterations'])

        self.assertTrue(count > 2 and count < 8)
        self.assertEqual(sum, 10)

    def test_loopmonitor_target_loop(self):
        target_cmd = [ './ci_test_macros', '0', 'none', '10' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'loop_monitor,trace,report',
            'CALI_LOOP_MONITOR_ITERATION_INTERVAL' : '5',
            'CALI_LOOP_MONITOR_TARGET_LOOPS'       : 'fooloop',
            'CALI_REPORT_CONFIG'     : 'select * where iteration#fooloop format expand',
            'CALI_REPORT_FILENAME'   : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test(target_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output[0])

        self.assertEqual(len(snapshots), 20)
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'iteration#mainloop' : '3',
                         'iteration#fooloop'  : '4',
                         'loop.iterations'    : '5',
                         'loop'               : 'mainloop/fooloop'
            }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'iteration#mainloop' : '7',
                         'iteration#fooloop'  : '9',
                         'loop.iterations'    : '5',
                         'loop'               : 'mainloop/fooloop'
            }))

    def test_regionmonitor_time_interval(self):
        target_cmd = [ './ci_test_macros', '5000', 'none', '5' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'region_monitor,trace,report',
            'CALI_REGION_MONITOR_TIME_INTERVAL' : '0.01',
            'CALI_REPORT_CONFIG'     : 'select * format expand',
            'CALI_REPORT_FILENAME'   : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test(target_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output[0])

        self.assertTrue(len(snapshots) == 4)
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'loop'     : 'mainloop/fooloop',
                         'function' : 'main/foo'
            }))
        self.assertFalse(cat.has_snapshot_with_attributes(
            snapshots, { 'annotation' : 'pre-loop' }))
        self.assertFalse(cat.has_snapshot_with_attributes(
            snapshots, { 'annotation' : 'before_loop' }))

if __name__ == "__main__":
    unittest.main()



#!@PYTHON_EXECUTABLE@

import os
import json
import time

nfib = 33

#
#   CONFIG ENV BEFORE LOADING TIMEMORY
#
if int(os.environ.get("EX_SAMPLE_NO_ENV", "0")) < 1:
    clinit = os.environ.get("TIMEMORY_COMPONENT_LIST_INIT", "").split()
    events = os.environ.get("TIMEMORY_CUPTI_EVENTS", "").split()
    metrics = os.environ.get("TIMEMORY_CUPTI_METRICS", "").split()

    clinit += ["cupti_counters"]
    events += ["active_warps", "global_load"]
    metrics += ["achieved_occupancy", "global_store"]

    os.environ["TIMEMORY_COMPONENT_LIST_INIT"] = "{}".format(",".join(clinit))
    os.environ["TIMEMORY_CUPTI_EVENTS"] = "{}".format(",".join(events))
    os.environ["TIMEMORY_CUPTI_METRICS"] = "{}".format(",".join(metrics))

import timemory  # noqa: E402


#
#    RUN FIBONACCI CALCULATION
#
def fibonacci(n):
    """
    Use this function to get CPU usage
    """
    return n if n < 2 else fibonacci(n - 1) + fibonacci(n - 2)


#
#   CONFIG
#
def get_timemory_config(config=""):
    """
    Provides auto-tuple configurations
    """
    configs = {}
    configs[None] = ["wall_clock", "cpu_clock"]
    items = configs[None] if config not in configs else configs[config]
    #
    # generate the list of enums
    #
    return [getattr(timemory.component, x) for x in items]


#
#   AUTO-TIMER
#
@timemory.bundle.auto_timer()
def foo():
    """
    Demonstrate decorator and context-manager with auto_timer
    Sleep for 2 seconds then run fibonacci calculation within context-manager
    """
    time.sleep(2)
    with timemory.bundle.auto_timer(key="[fibonacci]"):
        print("fibonacci({}) = {}".format(nfib, fibonacci(nfib)))


#
#   AUTO-TUPLE
#
@timemory.bundle.auto_tuple(get_timemory_config(), key="")
def bar():
    """
    Demonstrate decorator and context-manager with auto_tuple
    Run fibonacci calculation and then sleep for 2 seconds within context-manager
    """
    print("fibonacci({}) = {}".format(nfib, fibonacci(nfib)))
    with timemory.bundle.auto_tuple(get_timemory_config(), key="[sleep]"):
        time.sleep(2)


#
#   EXECUTE
#
if __name__ == "__main__":

    # do some work
    foo()
    bar()

    # get the dictionary
    data = timemory.get()
    # print the dictionary
    print(
        "TIMEMORY_DATA:\n{}".format(json.dumps(data, indent=4, sort_keys=True))
    )
    timemory.finalize()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

__version__ = '3.0.2'

try:
    import cPickle as pickle
except ImportError:
    import pickle

import functools
import inspect
import linecache
import optparse
import os
import sys

from ._line_profiler import LineProfiler as CLineProfiler

# Python 2/3 compatibility utils
# ===========================================================
PY3 = sys.version_info[0] == 3
PY35 = PY3 and sys.version_info[1] >= 5

# exec (from https://bitbucket.org/gutworth/six/):
if PY3:
    import builtins
    exec_ = getattr(builtins, "exec")
    del builtins
else:
    def exec_(_code_, _globs_=None, _locs_=None):
        """Execute code in a namespace."""
        if _globs_ is None:
            frame = sys._getframe(1)
            _globs_ = frame.f_globals
            if _locs_ is None:
                _locs_ = frame.f_locals
            del frame
        elif _locs_ is None:
            _locs_ = _globs_
        exec("""exec _code_ in _globs_, _locs_""")

if PY35:
    import inspect

    def is_coroutine(f):
        return inspect.iscoroutinefunction(f)
else:
    def is_coroutine(f):
        return False


# ============================================================
CO_GENERATOR = 0x0020


def is_generator(f):
    """ Return True if a function is a generator.
    """
    isgen = (f.__code__.co_flags & CO_GENERATOR) != 0
    return isgen


class LineProfiler(CLineProfiler):
    """ A profiler that records a metric on individual lines.
    """

    def __call__(self, func):
        """ Decorate a function to start the profiler on function entry and stop
        it on function exit.
        """
        self.add_function(func)
        if is_coroutine(func):
            wrapper = self.wrap_coroutine(func)
        elif is_generator(func):
            wrapper = self.wrap_generator(func)
        else:
            wrapper = self.wrap_function(func)
        return wrapper

    def wrap_generator(self, func):
        """ Wrap a generator to profile it.
        """
        @functools.wraps(func)
        def wrapper(*args, **kwds):
            g = func(*args, **kwds)
            # The first iterate will not be a .send()
            self.enable_by_count()
            try:
                item = next(g)
            except StopIteration:
                return
            finally:
                self.disable_by_count()
            input = (yield item)
            # But any following one might be.
            while True:
                self.enable_by_count()
                try:
                    item = g.send(input)
                except StopIteration:
                    return
                finally:
                    self.disable_by_count()
                input = (yield item)
        return wrapper

    def wrap_function(self, func):
        """ Wrap a function to profile it.
        """
        @functools.wraps(func)
        def wrapper(*args, **kwds):
            self.enable_by_count()
            try:
                result = func(*args, **kwds)
            finally:
                self.disable_by_count()
            return result
        return wrapper

    if PY35:
        from . import line_profiler_py35
        wrap_coroutine = line_profiler_py35.wrap_coroutine

    def dump_stats(self, filename):
        """ Dump a representation of the data to a file as a pickled LineStats
        object from `get_stats()`.
        """
        lstats = self.get_stats()
        with open(filename, 'wb') as f:
            pickle.dump(lstats, f, pickle.HIGHEST_PROTOCOL)

    def print_stats(self, stream=None, output_unit=None, stripzeros=False):
        """ Show the gathered statistics.
        """
        lstats = self.get_stats()
        show_text(lstats.metrics, lstats.unit, lstats.display_unit,
                  output_unit=output_unit, stream=stream, stripzeros=stripzeros)

    def run(self, cmd):
        """ Profile a single executable statment in the main namespace.
        """
        import __main__
        main_dict = __main__.__dict__
        return self.runctx(cmd, main_dict, main_dict)

    def runctx(self, cmd, globals, locals):
        """ Profile a single executable statement in the given namespaces.
        """
        self.enable_by_count()
        try:
            exec_(cmd, globals, locals)
        finally:
            self.disable_by_count()
        return self

    def runcall(self, func, *args, **kw):
        """ Profile a single function call.
        """
        self.enable_by_count()
        try:
            return func(*args, **kw)
        finally:
            self.disable_by_count()

    def add_module(self, mod):
        """ Add all the functions in a module and its classes.
        """
        from inspect import isclass, isfunction

        nfuncsadded = 0
        for item in mod.__dict__.values():
            if isclass(item):
                for k, v in item.__dict__.items():
                    if isfunction(v):
                        self.add_function(v)
                        nfuncsadded += 1
            elif isfunction(item):
                self.add_function(item)
                nfuncsadded += 1

        return nfuncsadded


def show_func(filename, start_lineno, func_name, metrics, unit,
              display_unit, output_unit=None, stream=None, stripzeros=False):
    """ Show results for a single function.
    """
    if stream is None:
        stream = sys.stdout

    template = '%6s %9s %12s %8s %8s  %-s'
    d = {}
    total_metric = 0.0
    linenos = []
    for lineno, nhits, metric in metrics:
        total_metric += metric
        linenos.append(lineno)

    if stripzeros and total_metric == 0:
        return

    if output_unit is None:
        output_unit = unit
    scalar = unit / output_unit

    stream.write("Total metric: %g %s\n" % (total_metric, display_unit))
    if os.path.exists(filename) or filename.startswith("<ipython-input-"):
        stream.write("File: %s\n" % filename)
        stream.write("Function: %s at line %s\n" % (func_name, start_lineno))
        if os.path.exists(filename):
            # Clear the cache to ensure that we get up-to-date results.
            linecache.clearcache()
        all_lines = linecache.getlines(filename)
        sublines = inspect.getblock(all_lines[start_lineno-1:])
    else:
        stream.write("\n")
        stream.write("Could not find file %s\n" % filename)
        stream.write(
            "Are you sure you are running this program from the same directory\n")
        stream.write("that you ran the profiler from?\n")
        stream.write("Continuing without the function's contents.\n")
        # Fake empty lines so we can see the metrics, if not the code.
        nlines = max(linenos) - min(min(linenos), start_lineno) + 1
        sublines = [''] * nlines
    for lineno, nhits, metric in metrics:
        perc = 0.0 if not total_metric > 0 else (100 * metric / total_metric)
        d[lineno] = (nhits,
                     '%5.1f' % (metric * scalar),
                     '%5.1f' % (float(metric) * scalar / nhits),
                     '%5.1f' % (perc))
    linenos = range(start_lineno, start_lineno + len(sublines))
    empty = ('', '', '', '')
    header = template % ('Line #', 'Hits', 'Metric', 'Per Hit', '% Metric',
                         'Line Contents')
    stream.write("\n")
    stream.write(header)
    stream.write("\n")
    stream.write('=' * len(header))
    stream.write("\n")
    for lineno, line in zip(linenos, sublines):
        nhits, metric, per_hit, percent = d.get(lineno, empty)
        txt = template % (lineno, nhits, metric, per_hit, percent,
                          line.rstrip('\n').rstrip('\r'))
        stream.write(txt)
        stream.write("\n")
    stream.write("\n")


def show_text(stats, unit, display_unit, output_unit=None, stream=None, stripzeros=False):
    """ Show text for the given metrics.
    """
    if stream is None:
        stream = sys.stdout

    if output_unit is not None:
        stream.write('Metric unit: %g (%s)\n\n' % (output_unit, display_unit))
    else:
        stream.write('Metric unit: %g %s\n\n' % (unit, display_unit))

    for (fn, lineno, name), metrics in sorted(stats.items()):
        show_func(fn, lineno, name, stats[fn, lineno, name], unit, display_unit,
                  output_unit=output_unit, stream=stream, stripzeros=stripzeros)


def load_ipython_extension(ip):
    """ API for IPython to recognize this module as an IPython extension.
    """
    try:
        from .line_profiler_ipython import LineProfilerMagics
        ip.register_magics(LineProfilerMagics)
    except Exception as e:
        print("line_profiler ipython extension import failed")


def load_stats(filename):
    """ Utility function to load a pickled LineStats object from a given
    filename.
    """
    with open(filename, 'rb') as f:
        return pickle.load(f)


def main():
    usage = "usage: python -m timemory.line_profiler <script>"

    parser = optparse.OptionParser(usage=usage,
                                   version=__version__)

    options, args = parser.parse_args()
    if len(args) != 1:
        parser.error("Must provide a filename.")
    lstats = load_stats(args[0])
    show_text(lstats.metrics, lstats.unit, lstats.display_unit)


if __name__ == '__main__':
    main()



# callpath service tests

import unittest

import calipertest as cat

class CaliperCallpathTest(unittest.TestCase):
    """ Caliper callpath smoketest """

    def test_callpath(self):
        target_cmd = [ './ci_test_alloc' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'callpath,trace,recorder',
            'CALI_CALLPATH_USE_NAME' : 'true',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) == 3)

        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, { 'callpath.address', 'callpath.regname', 'test_alloc.allocated.0', 'function' }))

        sreg = cat.get_snapshot_with_keys(snapshots, { 'callpath.regname', 'test_alloc.allocated.0' })

        self.assertTrue('main' in sreg.get('callpath.regname'))

if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2020, Lawrence Livermore National Security, LLC.
# See top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

from .caliperstreamreader import CaliperStreamReader
from .metadatadb import Attribute, Node, MetadataDB


class CaliperReader:
    """ Reads a Caliper .cali file and exports its contents.

    Use the read() method to read a Caliper .cali file. Contents are then
    available in the "records" and "globals" attributes. Each CaliperReader
    object should read a single file.

    Attributes:
        records : list-of-dicts
            The snapshot records read from the .cali file.
        globals : dict
            The global key:value attributes defined in the .cali file
    """

    def __init__(self):
        self.db = MetadataDB()

        self.records = []
        self.globals = {}

        self.num_records = {}


    def read(self, filename_or_stream):
        """ Read a .cali file or stream.

        After reading, the file's contents are available in the CaliperReader
        object's "records" and "globals" attributes.

        Arguments:
            filename_or_stream : str or stream object
                The filename or stream to read.
        """

        sr = CaliperStreamReader()

        sr.read(filename_or_stream, self._add_record)

        self.db = sr.db
        self.globals = sr.globals
        self.num_records = sr.num_records


    def attributes(self):
        """ Return the attribute keys.

        A file must have been read with read().

        Returns:
            The attribute keys (list of strings).
        """
        return self.db.attributes.keys()


    def attribute(self, attribute_name):
        """ Return a Caliper Attribute object for the given attribute name.
        """

        return self.db.attributes[attribute_name]


    def _add_record(self, rec):
        self.records.append(rec)


def read_caliper_contents(filename_or_stream):
    """ Reads a Caliper .cali file and returns its contents.

    Arguments:
        filename_or_stream:
            Stream object or the filename of the .cali file.

    Returns:
        (records, globals):
            A tuple with the Caliper snapshot records and globals. The first
            element contains the snapshot record as a list of dicts. The
            second element contains the global key:value attributes as a
            dict.
    """

    reader = CaliperReader()
    reader.read(filename_or_stream)

    return (reader.records, reader.globals)


def read_caliper_globals(filename_or_stream):
    """ Reads a Caliper .cali file and returns its global attributes.

    Arguments:
        filename_or_stream:
            Stream object or the filename of the .cali file.

    Returns:
        dict:
            The global key:value attributes as a dict.
    """

    reader = CaliperStreamReader()
    reader.read(filename_or_stream, None)

    return reader.globals


# Tests of the alloc service

import json, unittest

import calipertest as cat

class CaliperAllocServiceTest(unittest.TestCase):
    """ Caliper sysalloc service test cases """

    def helper_test_hook(self, hook):

        # test allocated.0

        self.assertTrue(cat.has_snapshot_with_keys(
            self.snapshots, { 'test_alloc.' + hook + '_hook',
                              'test_alloc.allocated.0',
                              'alloc.uid#ptr_in',
                              'ptr_in', 'ptr_out' }))

        self.assertTrue(cat.has_snapshot_with_attributes(
            self.snapshots, {'test_alloc.' + hook + '_hook' : 'true',
                             'test_alloc.allocated.0' : 'true',
                             'alloc.index#ptr_in'     : '0'
                             }))

        self.assertFalse(cat.has_snapshot_with_keys(
            self.snapshots, { 'test_alloc.' + hook + '_hook',
                             'test_alloc.allocated.0',
                             'alloc.uid#ptr_out' }))

        # test allocated.1

        self.assertTrue(cat.has_snapshot_with_keys(
            self.snapshots, { 'test_alloc.' + hook + '_hook',
                             'test_alloc.allocated.1',
                             'alloc.uid#ptr_in',
                             'ptr_in', 'ptr_out' }))

        index = '41' if hook == 'calloc' else '167'  # 41*sizeof(int)-1
        self.assertTrue(cat.has_snapshot_with_attributes(
            self.snapshots, {'test_alloc.' + hook + '_hook' : 'true',
                             'test_alloc.allocated.1' : 'true',
                             'alloc.index#ptr_in'     : index
                             }))

        self.assertFalse(cat.has_snapshot_with_keys(
            self.snapshots, { 'test_alloc.' + hook + '_hook',
                             'test_alloc.allocated.1',
                             'alloc.uid#ptr_out' }))

        # test allocated.freed

        self.assertTrue(cat.has_snapshot_with_keys(
            self.snapshots, { 'test_alloc.' + hook + '_hook',
                             'test_alloc.freed',
                             'ptr_in', 'ptr_out' }))

        self.assertFalse(cat.has_snapshot_with_keys(
            self.snapshots, { 'test_alloc.' + hook + '_hook',
                             'test_alloc.freed',
                             'alloc.uid#ptr_in' }))


    def test_alloc_hooks(self):
        target_cmd = [ './ci_test_alloc_hooks' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_ALLOC_RESOLVE_ADDRESSES' : 'true',
            'CALI_SERVICES_ENABLE'         : 'alloc:recorder:sysalloc:trace',
            'CALI_RECORDER_FILENAME'       : 'stdout',
            'CALI_LOG_VERBOSITY'           : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        self.snapshots = cat.get_snapshots_from_text(query_output)

        self.helper_test_hook('malloc')
        self.helper_test_hook('calloc')
        self.helper_test_hook('realloc')


    def test_mem_highwatermark_option(self):
        target_cmd = [ './ci_test_macros', '10', 'hatchet-region-profile,use.mpi=false,output=stdout,output.format=json,mem.highwatermark' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'     : '0'
        }

        obj = json.loads( cat.run_test(target_cmd, caliper_config)[0] )

        self.assertEqual(obj[1]['path'], 'main')
        self.assertTrue(float(obj[1]['Allocated MB']) > 0.0)

if __name__ == "__main__":
    unittest.main()



from os.path import join


def test_cli():
    """
    Test command line interaction with kernprof and line_profiler.

    References:
        https://github.com/pyutils/line_profiler/issues/9

    CommandLine:
        xdoctest -m ~/code/line_profiler/tests/test_cli.py test_cli
    """

    # Create a dummy source file
    import ubelt as ub
    code = ub.codeblock(
        '''
        @profile
        def my_inefficient_function():
            a = 0
            for i in range(10):
                a += i
                for j in range(10):
                    a += j

        if __name__ == '__main__':
            my_inefficient_function()
        ''')
    import tempfile
    tmp_dpath = tempfile.mkdtemp()
    tmp_src_fpath = join(tmp_dpath, 'foo.py')
    ub.writeto(tmp_src_fpath, code)

    # Run kernprof on it
    info = ub.cmd('kernprof -l {}'.format(tmp_src_fpath), verbose=3,
                  cwd=tmp_dpath)
    assert info['ret'] == 0

    tmp_lprof_fpath = join(tmp_dpath, 'foo.py.lprof')
    tmp_lprof_fpath

    info = ub.cmd('python -m line_profiler {}'.format(tmp_lprof_fpath),
                  cwd=tmp_dpath, verbose=3)
    assert info['ret'] == 0
    # Check for some patterns that should be in the output
    assert '% Time' in info['out']
    assert '7       100' in info['out']



#!/usr/bin/env python

import sys
import argparse
# read in components and mangled enums
from timemory_types import components, mangled_enums


def generate_case_label(component, func, indent_tabs=3, spaces=4, reference=True, template=True):
    """
    This function generates a case label for C++
    """
    enumeration = mangled_enums.get(component, component)

    spacer = " "*spaces             # a spacer of spaces length
    atab = "{}".format(spacer)      # the generic tab
    ftab = atab*indent_tabs         # the first tab level

    init_str = "." if reference else "->"
    if template:
        init_str += "template "
    return "{2}case {1}: obj{3}{4}<{0}>(); break;\n".format(
        component, enumeration.upper(), ftab, init_str, func)


if __name__ == "__main__":

    rbegin = "// GENERATE_SWITCH_REPLACE_BEGIN"
    rend = "// GENERATE_SWITCH_REPLACE_END"

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-a", "--additional", help="Additional components to produce case labels for", nargs='*')
    parser.add_argument(
        "-i", "--input", help="input file for replacing", type=str, default=None)
    parser.add_argument(
        "-o", "--output", help="output components from file", type=str, default=None)
    parser.add_argument("-e", "--exclude",
                        help="exclude components", action='store_true')
    parser.add_argument("-s", "--spaces-per-tab",
                        help="Number of spaces in a tab", type=int, default=4)
    parser.add_argument("-t", "--tabs-per-indent",
                        help="Number of tabs to indent", type=int, default=3)
    parser.add_argument("-r", "--replace", action='store_true',
                        help="Replace area between '{}' and '{}' in the output file".format(rbegin, rend))
    parser.add_argument(
        "-c", "--config", help="alternative to -i/--input when replacing", type=str, default=None)
    parser.add_argument("-P", "--pointer", help="Object operating on is pointer ('->') instead of reference ('.')",
                        action='store_true')
    parser.add_argument("-S", "--specialized", help="Initialization is specialized (not a template)",
                        action='store_true')
    parser.add_argument("-V", "--verbose",
                        help="Enable verbosity", default=0, type=int)
    parser.add_argument("-f", "--function",
                        help="Function name", default="init", type=str)

    args = parser.parse_args()

    if args.replace and args.input is None and args.output is None:
        raise Exception(
            "-r/--replace requires specifying -o/--output or -i/--input")

    # substitution string for replacing
    subdata = None

    if args.input is not None:
        with open(args.input, 'r') as f:
            subdata = f.read()

    if args.config is not None:
        with open(args.config, 'r') as f:
            components = f.read().strip().split()
        print("timemory components: [{}]\n".format(
            ", ".join(components)))

    if args.additional:
        components += args.additional

    if subdata is None and args.replace:
        if args.input is None:
            print("reading '{}' for replacement...".format(args.output))
            with open(args.output, 'r') as f:
                subdata = f.read()

    components.sort()
    if args.verbose > 0:
        print("timemory components: [{}]\n".format(", ".join(components)))
    outdata = ""
    for component in components:
        outdata += "{}".format(generate_case_label(component, args.function,
                                                   args.tabs_per_indent, args.spaces_per_tab,
                                                   not args.pointer, not args.specialized))
    outdata = outdata.strip("\n")
    spacer = " "*args.spaces_per_tab    # a spacer of spaces length
    atab = "{}".format(spacer)          # the generic tab
    ftab = atab*args.tabs_per_indent    # the first tab level

    outdata += "\n{}case TIMEMORY_COMPONENTS_END:\n{}default: break;".format(
        ftab, ftab)

    if subdata is not None:
        try:
            sbegin = subdata.find(rbegin) + len(rbegin) + 1
            send = subdata.find(rend)
            substr = subdata[sbegin:send]
            outdata += " " * args.spaces_per_tab * (args.tabs_per_indent + 1)
            if args.verbose > 1:
                print("sbegin = {}\nsend = {}\nsubstring:\n{}\n\noutdata:\n{}\n".format(
                      sbegin, send, substr, outdata))
            outdata = subdata.replace(substr, outdata)
            if args.verbose > 1:
                print("converted:\n{}\n".format(subdata))
        except Exception as e:
            print(e)
            raise

    if args.output is not None:
        with open(args.output, 'w') as f:
            f.write(outdata)
    else:
        print("{}".format(outdata))



#!@PYTHON_EXECUTABLE@
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import

__author__ = "Jonathan Madsen"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Jonathan Madsen"]
__license__ = "MIT"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"

try:
    import mpi4py  # noqa: F401
    from mpi4py import MPI  # noqa: F401
except ImportError:
    pass

import json
import unittest
import timemory as tim


# --------------------------- Hatchet Tests set ---------------------------------------- #
# Hatchet tests class
class TimemoryToolsTests(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # set up environment variables
        tim.settings.verbose = 1
        tim.settings.debug = False
        tim.settings.tree_output = True
        tim.settings.text_output = True
        tim.settings.cout_output = False
        tim.settings.json_output = False
        tim.settings.flamegraph_output = False
        tim.settings.mpi_thread = False
        tim.settings.dart_output = True
        tim.settings.dart_count = 1
        tim.settings.banner = False
        tim.settings.memory_units = "MB"

    def setUp(self):
        pass

    def tearDown(self):
        pass

    @classmethod
    def tearDownClass(self):
        pass

    # ---------------------------------------------------------------------------------- #
    # test handling wall-clock data from hatchet
    def test_mallocp(self):
        """mallocp"""

        try:
            import numpy as np
            import gc
        except ImportError:
            return

        if not tim.component.is_available("malloc_gotcha"):
            return

        from timemory.component import MallocGotcha
        from timemory.storage import MallocGotchaStorage

        _idx = tim.start_mallocp()
        _arr = np.ones([1000, 1000], dtype=np.float64)
        _sum = np.sum(_arr)
        del _arr
        gc.collect()
        _idx = tim.stop_mallocp(_idx)

        self.assertTrue(_sum == 1000 * 1000)
        self.assertTrue(_idx == 0)

        _json = tim.get(components=[MallocGotcha.id()])
        _data = MallocGotchaStorage.get()
        _alloc = 0.0
        _dealloc = 0.0

        print("{}\n".format(json.dumps(_json, indent=4)))

        for itr in _data:
            if "malloc" in itr.prefix() or "calloc" in itr.prefix():
                _alloc += itr.data().get_raw()
            if "free" in itr.prefix():
                _dealloc += itr.data().get_raw()

        _minsz = 1000 * 1000 * 8
        print(f"Minimum size     : {_minsz}")
        print(f"Allocated size   : {_alloc}")
        print(f"Deallocated size : {_alloc}")

        self.assertTrue(_alloc > _minsz)

        unit_v = _json["timemory"]["malloc_gotcha"]["unit_value"]
        unit_r = _json["timemory"]["malloc_gotcha"]["unit_repr"]
        real_v = MallocGotcha.unit()
        real_r = MallocGotcha.display_unit().upper()

        self.assertEqual(unit_v, real_v)
        self.assertEqual(unit_r, real_r)


# ----------------------------- main test runner -------------------------------------- #
# main runner
def run():
    # run all tests
    unittest.main()


if __name__ == "__main__":
    tim.initialize([__file__])
    run()



#!/usr/bin/env python

import sys
import argparse
# read in components and mangled enums
from timemory_types import components, mangled_enums, recommended_types


def generate_list(component, indent_tabs=3, spaces=4, reference=True, template=True):
    """
    This function generates a case label for C++
    """
    enumeration = mangled_enums.get(component, component)

    spacer = " "*spaces           # a spacer of spaces length
    atab = "{}".format(spacer)  # the generic tab
    ftab = atab*indent_tabs     # the first tab level
    stab = atab*(indent_tabs+1)  # the second tab level

    init_str = "." if reference else "->"
    if template:
        init_str += "template "
    return "{2}case {1}: obj{3}init<{0}>(); break;\n".format(
        component, enumeration.upper(), ftab, init_str)


if __name__ == "__main__":

    rbegin = "// GENERATE_SWITCH_REPLACE_BEGIN"
    rend = "// GENERATE_SWITCH_REPLACE_END"

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-a", "--additional", help="Additional components to produce case labels for", nargs='*')
    parser.add_argument(
        "-i", "--input", help="input file for replacing", type=str, default=None)
    parser.add_argument(
        "-o", "--output", help="output components from file", type=str, default=None)
    parser.add_argument("-e", "--exclude",
                        help="exclude components", action='store_true')
    parser.add_argument("-s", "--spaces-per-tab",
                        help="Number of spaces in a tab", type=int, default=4)
    parser.add_argument("-t", "--tabs-per-indent",
                        help="Number of tabs to indent", type=int, default=3)
    parser.add_argument("-r", "--replace", action='store_true',
                        help="Replace area between '{}' and '{}' in the output file".format(rbegin, rend))
    parser.add_argument(
        "-c", "--config", help="alternative to -i/--input when replacing", type=str, default=None)
    parser.add_argument("-P", "--pointer", help="Object operating on is pointer ('->') instead of reference ('.')",
                        action='store_true')
    parser.add_argument("-S", "--specialized", help="Initialization is specialized (not a template)",
                        action='store_true')
    parser.add_argument("-V", "--verbose",
                        help="Enable verbosity", default=0, type=int)

    args = parser.parse_args()

    if args.replace and args.input is None and args.output is None:
        raise Exception(
            "-r/--replace requires specifying -o/--output or -i/--input")

    # substitution string for replacing
    subdata = None

    if args.input is not None:
        with open(args.input, 'r') as f:
            subdata = f.read()

    if args.config is not None:
        with open(args.config, 'r') as f:
            components = f.read().strip().split()
        print("timemory components: [{}]\n".format(
            ", ".join(components)))

    if args.additional:
        components += args.additional

    if subdata is None and args.replace:
        if args.input is None:
            print("reading '{}' for replacement...".format(args.output))
            with open(args.output, 'r') as f:
                subdata = f.read()

    components.sort()
    if args.verbose > 0:
        print("timemory components: [{}]\n".format(", ".join(components)))

    #
    outdata = "using complete_tuple_t = std::tuple<\n"
    for component in components:
        outdata += "\tcomponent::{},\n".format(component)
    outdata = outdata.strip("\n")
    outdata = outdata.strip(",")
    outdata += ">;\n\n"

    outdata += "using complete_auto_list_t = auto_list<\n"
    for component in components:
        outdata += "\tcomponent::{},\n".format(component)
    outdata = outdata.strip("\n")
    outdata = outdata.strip(",")
    outdata += ">;\n\n"
    outdata += "using complete_list_t = complete_auto_list_t::component_type;\n\n"

    a_hybdata = "using recommended_auto_hybrid_t = auto_hybrid<\n"
    c_hybdata = "using recommended_hybrid_t = component_hybrid<\n"
    for key, items in recommended_types.items():
        outdata += "using recommended_auto_{0}_t = auto_{0}<\n".format(key)
        for item in items:
            outdata += "\tcomponent::{},\n".format(item)
        outdata = outdata.strip("\n")
        outdata = outdata.strip(",")
        outdata += ">;\n\n"
        outdata += "using recommended_{0}_t = recommended_auto_{0}_t::component_type;\n\n".format(
            key)
        a_hybdata += "\trecommended_{0}_t,\n".format(key)
        c_hybdata += "\trecommended_{0}_t,\n".format(key)
    a_hybdata = a_hybdata.strip("\n")
    a_hybdata = a_hybdata.strip(",")
    a_hybdata += ">;\n\n"
    c_hybdata = c_hybdata.strip("\n")
    c_hybdata = c_hybdata.strip(",")
    c_hybdata += ">;\n\n"

    outdata += a_hybdata
    outdata += c_hybdata

    if subdata is not None:
        try:
            sbegin = subdata.find(rbegin) + len(rbegin) + 1
            send = subdata.find(rend)
            substr = subdata[sbegin:send]
            outdata += " " * args.spaces_per_tab * (args.tabs_per_indent + 1)
            if args.verbose > 1:
                print("sbegin = {}\nsend = {}\nsubstring:\n{}\n\noutdata:\n{}\n".format(
                      sbegin, send, substr, outdata))
            outdata = subdata.replace(substr, outdata)
            if args.verbose > 1:
                print("converted:\n{}\n".format(subdata))
        except Exception as e:
            print(e)
            raise

    if args.output is not None:
        with open(args.output, 'w') as f:
            f.write(outdata)
    else:
        print("{}".format(outdata))



# Copyright (c) 2020, Lawrence Livermore National Security, LLC.
# See top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

class Node:
    """ A Caliper metadata tree node. """

    CALI_INV_ID = 0xFFFFFFFFFFFFFFFF

    def __init__(self, metadb, id, attribute_id, data):
        self.parent       = None
        self.children     = []

        self.id           = id
        self.attribute_id = attribute_id
        self.data         = data

        self.metadb       = metadb


    def append(self, child):
        """ Append a child node to this node.
        """

        child.parent = self
        self.children.append(child)

    def get(self, attribute_id_or_name):
        """ Iterate through parents and return the data element for the first
        node with the given attribute, or None when no such element exists.
        """

        if isinstance(attribute_id_or_name, str):
            if not attribute_id_or_name in self.metadb.attributes:
                return None

            attribute_id = self.metadb.attributes[attribute_id_or_name].id()
        else:
            attribute_id = attribute_id_or_name

        node = self
        while node is not None and node.attribute_id != attribute_id:
            node = node.parent

        return node.data if node else None

    def attribute(self):
        """ Return the Caliper attribute object for this node.
        """
        return Attribute(self.metadb.nodes[self.attribute_id])


class Attribute:
    """ A Caliper attribute key.

    Provides information about Caliper attribute keys.
    """

    attr_attribute_id =   8
    type_attribute_id =   9
    prop_attribute_id =  10

    CALI_ATTR_ASVALUE =   1
    CALI_ATTR_NESTED  = 256

    def __init__(self, node):
        self.node = node

    def name(self):
        """ Return the name of the Caliper attribute. """
        return self.node.data

    def id(self):
        """ Return the node ID of the Caliper attribute. """
        return self.node.id

    def get(self, attribute):
        """ Return metadata value for this attribute.

        Arguments:
            attribute : string or integer
                Metadata attribute name or ID

        Returns:
            (object):
                The metadata value.
        """

        return self.node.get(attribute)

    def attribute_type(self):
        """ Return the name of the attribute's Caliper data type. """

        return self.node.get(self.type_attribute_id)[1]

    def is_nested(self):
        """ Does this attribute have the Caliper "nested" property?

        Attributes with the "nested" property define the hierarchy of Caliper
        begin/end annotations.
        """

        prop = self.node.get(self.prop_attribute_id)
        return (int(prop) & self.CALI_ATTR_NESTED) != 0

    def is_value(self):
        """ Does this attribute have the "as_value" property?

        Attributes with the "as_value" property are stored directly in
        snapshot records.
        """

        prop = self.node.get(self.prop_attribute_id)
        return (int(prop) & self.CALI_ATTR_ASVALUE) != 0


class MetadataDB:
    """ The Caliper metadata tree """

    def __init__(self):
        # The Caliper attribute type definitions
        self.types = [
            ( 0, 'inv'    ),
            ( 1, 'usr'    ),
            ( 2, 'int'    ),
            ( 3, 'uint'   ),
            ( 4, 'string' ),
            ( 5, 'addr'   ),
            ( 6, 'double' ),
            ( 7, 'bool'   ),
            ( 8, 'type'   ),
            ( 9, 'ptr'    )
        ]

        #   Initialize the node list with the pre-defined bootstrap nodes.
        # Use a dict for nodes because there can be gaps in the node list.
        self.nodes = {
             0: Node(self,  0, 9, self.types[1]),
             1: Node(self,  1, 9, self.types[2]),
             2: Node(self,  2, 9, self.types[3]),
             3: Node(self,  3, 9, self.types[4]),
             4: Node(self,  4, 9, self.types[5]),
             5: Node(self,  5, 9, self.types[6]),
             6: Node(self,  6, 9, self.types[7]),
             7: Node(self,  7, 9, self.types[8]),
             8: Node(self,  8, 8, 'cali.attribute.name'),
             9: Node(self,  9, 8, 'cali.attribute.type'),
            10: Node(self, 10, 8, 'cali.attribute.prop'),
            11: Node(self, 11, 9, self.types[9])
        }

        # Append bootstrap node children to parents
        self.nodes[3].append(self.nodes[ 8])
        self.nodes[7].append(self.nodes[ 9])
        self.nodes[1].append(self.nodes[10])

        # Initialize the attributes dict
        self.attributes = {
            'cali.attribute.name': Attribute(self.nodes[ 8]),
            'cali.attribute.type': Attribute(self.nodes[ 9]),
            'cali.attribute.prop': Attribute(self.nodes[10])
        }


    def import_node(self, node_id, attribute_id, data, parent_id = Node.CALI_INV_ID):
        node = Node(self, node_id, attribute_id, data)
        self.nodes[node_id] = node

        if parent_id in self.nodes:
            self.nodes[parent_id].append(node)

        if attribute_id == Attribute.attr_attribute_id:
            self.attributes[data] = Attribute(node)



# MPI application tests
# For simplicity, these tests only run on a single rank without mpiexec

import json
import os
import unittest

import calipertest as cat

class CaliperMPITest(unittest.TestCase):
    """ Test MPI init before Caliper init"""

    def test_mpi_before_cali(self):
        target_cmd = [ './ci_test_mpi_before_cali' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'PATH'                    : '/usr/bin', # for ssh/rsh
            'CALI_LOG_VERBOSITY'      : '0',
            'CALI_SERVICES_ENABLE'    : 'event,mpi,mpireport,trace',
            'CALI_MPI_WHITELIST'      : 'all',
            'CALI_MPIREPORT_FILENAME' : 'stdout',
            'CALI_MPIREPORT_CONFIG'   : 'select count() group by function,mpi.function,mpi.rank format cali'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function'       : 'main',
                         'mpi.function'   : 'MPI_Barrier',
                         'count'          : '2',
                         'mpi.rank'       : '0'
            }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function' : 'main', 'mpi.function' : 'MPI_Bcast',   'count' : '1' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function' : 'main', 'mpi.function' : 'MPI_Reduce',  'count' : '1' }))

    def test_cali_before_mpi(self):
        target_cmd = [ './ci_test_cali_before_mpi' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'PATH'                    : '/usr/bin', # for ssh/rsh
            'CALI_LOG_VERBOSITY'      : '0',
            'CALI_SERVICES_ENABLE'    : 'event,mpi,mpireport,trace',
            'CALI_MPI_WHITELIST'      : 'all',
            'CALI_MPIREPORT_FILENAME' : 'stdout',
            'CALI_MPIREPORT_CONFIG'   : 'select count() group by function,mpi.function,mpi.rank format cali'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function'       : 'main',
                         'mpi.function'   : 'MPI_Barrier',
                         'count'          : '2',
                         'mpi.rank'       : '0'
            }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function' : 'main', 'mpi.function' : 'MPI_Bcast',   'count' : '1' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function' : 'main', 'mpi.function' : 'MPI_Reduce',  'count' : '1' }))

    def test_mpi_whitelist(self):
        target_cmd = [ './ci_test_mpi_before_cali' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'PATH'                    : '/usr/bin', # for ssh/rsh
            'CALI_LOG_VERBOSITY'      : '0',
            'CALI_SERVICES_ENABLE'    : 'event,mpi,mpireport,trace',
            'CALI_MPI_WHITELIST'      : 'MPI_Bcast',
            'CALI_MPIREPORT_FILENAME' : 'stdout',
            'CALI_MPIREPORT_CONFIG'   : 'select count() group by function,mpi.function format cali'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertFalse(cat.has_snapshot_with_attributes(
            snapshots, { 'function' : 'main', 'mpi.function' : 'MPI_Barrier' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function' : 'main', 'mpi.function' : 'MPI_Bcast', 'count' : '1' }))
        self.assertFalse(cat.has_snapshot_with_attributes(
            snapshots, { 'function' : 'main', 'mpi.function' : 'MPI_Reduce'  }))

    def test_mpi_blacklist(self):
        target_cmd = [ './ci_test_mpi_before_cali' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'PATH'                    : '/usr/bin', # for ssh/rsh
            'CALI_LOG_VERBOSITY'      : '0',
            'CALI_SERVICES_ENABLE'    : 'event,mpi,mpireport,trace',
            'CALI_MPI_BLACKLIST'      : 'MPI_Barrier,MPI_Reduce',
            'CALI_MPIREPORT_FILENAME' : 'stdout',
            'CALI_MPIREPORT_CONFIG'   : 'select count() group by function,mpi.function format cali'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertFalse(cat.has_snapshot_with_attributes(
            snapshots, { 'function' : 'main', 'mpi.function' : 'MPI_Barrier' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function' : 'main', 'mpi.function' : 'MPI_Bcast', 'count' : '1' }))
        self.assertFalse(cat.has_snapshot_with_attributes(
            snapshots, { 'function' : 'main', 'mpi.function' : 'MPI_Reduce'  }))

    def test_mpi_whitelist_and_blacklist(self):
        target_cmd = [ './ci_test_mpi_before_cali' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'PATH'                    : '/usr/bin', # for ssh/rsh
            'CALI_LOG_VERBOSITY'      : '0',
            'CALI_SERVICES_ENABLE'    : 'event,mpi,mpireport,trace',
            'CALI_MPI_WHITELIST'      : 'MPI_Bcast,MPI_Reduce',
            'CALI_MPI_BLACKLIST'      : 'MPI_Reduce',
            'CALI_MPIREPORT_FILENAME' : 'stdout',
            'CALI_MPIREPORT_CONFIG'   : 'select count() group by function,mpi.function format cali'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertFalse(cat.has_snapshot_with_attributes(
            snapshots, { 'function' : 'main', 'mpi.function' : 'MPI_Barrier' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function' : 'main', 'mpi.function' : 'MPI_Bcast', 'count' : '1' }))
        self.assertFalse(cat.has_snapshot_with_attributes(
            snapshots, { 'function' : 'main', 'mpi.function' : 'MPI_Reduce'  }))

    def test_mpi_msg_trace(self):
        target_cmd = [ './ci_test_mpi_before_cali' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'PATH'                    : '/usr/bin', # for ssh/rsh
            'CALI_LOG_VERBOSITY'      : '0',
            'CALI_SERVICES_ENABLE'    : 'event,mpi,recorder,trace',
            'CALI_MPI_MSG_TRACING'    : 'true',
            'CALI_MPI_WHITELIST'      : 'all',
            'CALI_RECORDER_FILENAME'  : 'stdout'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function'          : 'main',
                         'mpi.function'      : 'MPI_Barrier',
                         'mpi.coll.type'     : '1',
                         'mpi.comm.is_world' : 'true'
            }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function'          : 'main',
                         'mpi.function'      : 'MPI_Bcast',
                         'mpi.coll.type'     : '3',
                         'mpi.coll.root'     : '0',
                         'mpi.comm.is_world' : 'true'
            }))
        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, { 'function', 'mpi.function', 'mpi.coll.type', 'mpi.call.id'
            }))

    def test_mpireport_controller(self):
        target_cmd = [ './ci_test_mpi_before_cali', 'mpi-report' ]

        caliper_config = {
            'PATH'                    : '/usr/bin', # for ssh/rsh
            'CALI_LOG_VERBOSITY'      : '0',
        }

        log_targets = [
            'Function',
            'MPI_Bcast'
        ]

        report_out,_ = cat.run_test(target_cmd, caliper_config)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

if __name__ == "__main__":
    unittest.main()



# Libpfm tests

import unittest

import calipertest as calitest

class CaliperLibpfmTest(unittest.TestCase):
    """ Caliper Libpfm test case """

    def test_libpfm_counting(self):
        target_cmd = [ './ci_test_basic' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'          : 'event:libpfm:trace:recorder',
            'CALI_LIBPFM_ENABLE_SAMPLING'   : 'false',
            'CALI_LIBPFM_RECORD_COUNTERS'   : 'true',
            'CALI_LIBPFM_EVENTS'            : 'cycles',
            'CALI_LIBPFM_SAMPLE_PERIOD'     : '10000000',
            'CALI_LIBPFM_SAMPLE_ATTRIBUTES' : 'ip,time',
            'CALI_RECORDER_FILENAME'        : 'stdout',
            'CALI_LOG_VERBOSITY'            : '0'
        }

        query_output = calitest.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = calitest.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) > 1)

        self.assertTrue(calitest.has_snapshot_with_keys(
            snapshots, { 'libpfm.counter.cycles', 'phase', 'iteration' }))

    def test_libpfm_sampling(self):
        target_cmd = [ './ci_test_macros', '50', 'none', '100' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'        : 'event:libpfm:pthread:trace:recorder',
            'CALI_LIBPFM_ENABLE_SAMPLING' : 'true',
            'CALI_LIBPFM_RECORD_COUNTERS' : 'false',
            'CALI_LIBPFM_EVENTS'          : 'instructions',
            'CALI_RECORDER_FILENAME'      : 'stdout',
            'CALI_LOG_VERBOSITY'          : '0'
        }

        query_output = calitest.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = calitest.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) > 1)

        self.assertTrue(calitest.has_snapshot_with_keys(
            snapshots, { 'libpfm.event_sample_name', 'libpfm.ip', 'libpfm.time' }))

if __name__ == "__main__":
    unittest.main()



# Basic smoke tests: aggregation

import unittest

import calipertest as calitest

class CaliperAggregationTest(unittest.TestCase):
    """ Caliper test case """

    def test_aggregate_default(self):
        target_cmd = [ './ci_test_aggregate' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'aggregate:event:recorder:timestamp',
            'CALI_TIMER_SNAPSHOT_DURATION' : 'true',
            'CALI_TIMER_INCLUSIVE_DURATION' : 'true',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = calitest.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = calitest.get_snapshots_from_text(query_output)

        self.assertTrue(calitest.has_snapshot_with_keys(
            snapshots, [ 'loop.id', 'function', 
                         'sum#time.inclusive.duration',
                         'min#time.inclusive.duration',
                         'max#time.inclusive.duration',
                         'sum#time.duration',
                         'count' ] ))

        self.assertTrue(calitest.has_snapshot_with_attributes(
            snapshots, {
                'event.end#function': 'foo', 
                'loop.id': 'A',
                'count': '6' }))
        self.assertTrue(calitest.has_snapshot_with_attributes(
            snapshots, {
                'event.end#function': 'foo', 
                'loop.id': 'B',
                'count': '4' }))

    def test_aggregate_combined_key(self):
        target_cmd = [ './ci_test_aggregate' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'aggregate:event:recorder',
            'CALI_AGGREGATE_KEY'     : 'event.end#function,iteration',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = calitest.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = calitest.get_snapshots_from_text(query_output)

        self.assertTrue(calitest.has_snapshot_with_attributes(
            snapshots, {
                'event.end#function': 'foo', 
                'iteration': '1',
                'count': '3' }))
        self.assertTrue(calitest.has_snapshot_with_attributes(
            snapshots, {
                'event.end#function': 'foo', 
                'iteration': '3',
                'count': '1' }))

    def test_aggregate_value_key(self):
        target_cmd = [ './ci_test_aggregate' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'aggregate:event:recorder',
            'CALI_AGGREGATE_KEY'     : 'iteration',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = calitest.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = calitest.get_snapshots_from_text(query_output)

        self.assertTrue(calitest.has_snapshot_with_attributes(
            snapshots, {
                'iteration': '1',
                'count': '8' }))
        self.assertTrue(calitest.has_snapshot_with_attributes(
            snapshots, {
                'iteration': '3',
                'count': '3' }))
        self.assertFalse(calitest.has_snapshot_with_keys(
            snapshots, [ 'function', 'loop.id' ]))

    def test_aggregate_nested(self):
        target_cmd = [ './ci_test_aggregate' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'aggregate:event:recorder',
            'CALI_AGGREGATE_KEY'     : 'prop:nested',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = calitest.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = calitest.get_snapshots_from_text(query_output)

        self.assertTrue(calitest.has_snapshot_with_attributes(
            snapshots, {
                'loop.id' : 'A',
                'function': 'foo',
                'count'   : '6' }))
        self.assertTrue(calitest.has_snapshot_with_attributes(
            snapshots, {
                'loop.id' : 'B',
                'function': 'foo',
                'count'   : '4' }))

    def test_aggregate_implicit_and_value(self):
        target_cmd = [ './ci_test_aggregate' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'aggregate:event:recorder',
            'CALI_AGGREGATE_KEY'     : '*,iteration',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = calitest.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = calitest.get_snapshots_from_text(query_output)

        self.assertTrue(calitest.has_snapshot_with_attributes(
            snapshots, {
                'loop.id'  : 'A',
                'event.end#function': 'foo',
                'iteration': '1',
                'count'    : '2' }))
        self.assertTrue(calitest.has_snapshot_with_attributes(
            snapshots, {
                'loop.id'  : 'B',
                'function' : 'foo',
                'iteration': '3',
                'count'    : '1' }))

    def test_aggregate_attributes(self):
        target_cmd = [ './ci_test_aggregate' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'aggregate:event:recorder:timestamp',
            'CALI_TIMER_SNAPSHOT_DURATION' : 'true',
            'CALI_TIMER_INCLUSIVE_DURATION' : 'true',
            'CALI_AGGREGATE_ATTRIBUTES' : 'time.duration',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = calitest.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = calitest.get_snapshots_from_text(query_output)

        self.assertTrue(calitest.has_snapshot_with_keys(
            snapshots, [ 'loop.id', 'function',
                         'sum#time.duration',
                         'count' ] ))
        self.assertFalse(calitest.has_snapshot_with_keys(
            snapshots, [ 'sum#time.inclusive.duration' ] ))

if __name__ == "__main__":
    unittest.main()



# Tests of the Fortran API

import unittest

import calipertest as cat

class CaliperFortranAPITest(unittest.TestCase):
    """ Caliper Fortran API test cases """

    def test_f_ann_trace(self):
        target_cmd = [ './ci_test_f_ann', 'event-trace,output=stdout' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, { 'annotation', 'time.inclusive.duration' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'annotation': 'main/work' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'annotation': 'main/foo' }))

    def test_f_ann_globals(self):
        target_cmd = [ './ci_test_f_ann', 'event-trace,output=stdout' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '--list-globals', '-e' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'global.int': '42' }))

if __name__ == "__main__":
    unittest.main()



"""
mkinit ~/code/line_profiler/line_profiler/__init__.py --relative
mkinit ~/code/line_profiler/line_profiler/__init__.py --relative -w
"""

__submodules__ = [
    'line_profiler', 'line_profiler_ipython'
]

from .line_profiler import __version__

try:
    from .line_profiler import (CO_GENERATOR, LineProfiler, LineProfilerMagics,
                                PY3, PY35, exec_, is_coroutine, is_generator,
                                load_ipython_extension, load_stats, main,
                                show_func, show_text,)

    __all__ = ['CO_GENERATOR', 'LineProfiler', 'LineProfilerMagics', 'PY3', 'PY35',
               'exec_', 'is_coroutine', 'is_generator', 'line_profiler',
               'load_ipython_extension', 'load_stats', 'main', 'show_func',
               'show_text', '__version__']
except:
    from .line_profiler import (CO_GENERATOR, LineProfiler,
                                PY3, PY35, exec_, is_coroutine, is_generator,
                                load_ipython_extension, load_stats, main,
                                show_func, show_text,)

    __all__ = ['CO_GENERATOR', 'LineProfiler', 'PY3', 'PY35',
               'exec_', 'is_coroutine', 'is_generator', 'line_profiler',
               'load_ipython_extension', 'load_stats', 'main', 'show_func',
               'show_text', '__version__']
